# coding=utf-8
# --------------------------------------------------------------------------------------------------------------------
# Licensed under the MIT LICENSE. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------------------------------
"""Main Bancor Simulation module interface."""

import mesa
import itertools
from bancor_research import Decimal
from mesa.datacollection import DataCollector
from pydantic.types import Any

from bancor_research.bancor_simulator.v3.simulation.agents import Trader, LP, Protocol

"""
If you want to perform a parameter sweep, call batch_run.py instead of run.py.
For details see batch_run.py in the same directory as run.py.
"""


def track_params(model):
    return model.init_traders, model.init_lps


def track_run(model):
    return model.uid


def get_bntkn_health(model):
    bntkn_health = [a.protocol.bntkn_health for a in model.schedule.agents][0]
    return int(round(float(bntkn_health)))


def get_bnbnt_health(model):
    bnbnt_health = [a.protocol.bnbnt_health for a in model.schedule.agents][0]
    return int(round(float(bnbnt_health)))


class BancorSimulation(mesa.Model):
    """
    This is a highly abstracted, simplified model of the Bancor v3 protocol, with only two
    types of agents (Trader and LP) and a single BancorDapp instance representing their environment and state space(s).
    Note that the LPs perform those actions which are in reality only performed vy the DAO, such as changing fees, etc..
    """

    current_id = 0
    id_gen = itertools.count(1)

    # grid height
    grid_h = 20
    # grid width
    grid_w = 20
    max_steps = SIMULATION_MAX_STEPS

    def __init__(
        self,
        height=grid_h,
        width=grid_w,
        init_traders=SIMULATION_NUM_TRADERS,
        init_lps=SIMULATION_NUM_LPs,
        whale_threshold=SIMULATION_WHALE_THRESHOLD,
        trading_fee=float(DEFAULT_TRADING_FEE),
        network_fee=float(DEFAULT_NETWORK_FEE),
        withdrawal_fee=float(DEFAULT_WITHDRAWAL_FEE),
        whitelisted_tokens=None,
        target_tvl=SIMULATION_TARGET_TVL,
        price_feeds=DEFAULT_PRICE_FEEDS,
        bnt_funding_limit=float(DEFAULT_BNT_FUNDING_LIMIT),
        bnt_min_liquidity=float(DEFAULT_BNT_MIN_LIQUIDITY),
        *args: Any,
        **kwargs: Any
    ):

        if whitelisted_tokens is None:
            whitelisted_tokens = SIMULATION_WHITELISTED_TOKENS

        # create a single BancorDapp v3 instance for the model
        self.protocol = Protocol(
            self.next_id(),
            self,
            Decimal(trading_fee),
            Decimal(network_fee),
            Decimal(withdrawal_fee),
            whitelisted_tokens,
            price_feeds,
            Decimal(bnt_funding_limit),
            Decimal(bnt_min_liquidity),
        )

        super().__init__(*args, **kwargs)

        self.height = height
        self.width = width
        self.init_lps = init_lps
        self.init_traders = init_traders
        self.uid = next(self.id_gen)
        self.schedule = mesa.time.RandomActivation(self)
        self.grid = mesa.space.MultiGrid(self.width, self.height, torus=True)
        self.whale_threshold = whale_threshold
        self.trading_fee = float(trading_fee)
        self.network_fee = float(network_fee)
        self.withdrawal_fee = float(withdrawal_fee)

        self.datacollector = DataCollector(
            model_reporters={
                "bnbnt_health": get_bnbnt_health,
                "bntkn_health": get_bntkn_health,
                "Model Params": track_params,
                "Run": track_run,
            },
            agent_reporters={
                "profit": "profit",
                "latest_user_name": "user_name",
                "latest_action": "latest_action",
                "latest_tkn_name": "latest_tkn_name",
                "latest_tkn_amt": "latest_tkn_amt",
                "master_vault_bnt": "master_vault_bnt",
                "erc20contracts_bntkn": "erc20contracts_bntkn",
                "staked_bnt": "staked_bnt",
                "external_protection_vault_tkn": "external_protection_vault_tkn",
                "spot_rate": "spot_rate",
                "ema_rate": "ema_rate",
                "protocol_wallet_bnbnt": "protocol_wallet_bnbnt",
                "vortex_bnt": "vortex_bnt",
                "erc20contracts_bnbnt": "erc20contracts_bnbnt",
                "master_vault_tkn": "master_vault_tkn",
                "staked_tkn": "staked_tkn",
            },
        )

        # create Traders for the model according to parameters
        for i in range(self.init_traders):
            x1 = self.random.randrange(self.width)
            y1 = self.random.randrange(self.height)
            trader = Trader(
                unique_id=self.next_id(),
                pos=(x1, y1),
                model=self,
                moore=True,
                whale_threshold=whale_threshold,
                whitelisted_tokens=whitelisted_tokens,
                target_tvl=target_tvl,
            )
            self.grid.place_agent(trader, (x1, y1))
            self.schedule.add(trader)

        # create LPs for the model according to parameters
        for i in range(self.init_lps):
            x2 = self.random.randrange(self.width)
            y2 = self.random.randrange(self.height)
            liquidity_provider = LP(
                unique_id=self.next_id(),
                pos=(x2, y2),
                model=self,
                moore=True,
                whale_threshold=whale_threshold,
                whitelisted_tokens=whitelisted_tokens,
                target_tvl=target_tvl,
            )
            self.grid.place_agent(liquidity_provider, (x2, y2))
            self.schedule.add(liquidity_provider)

        self.running = True
        self.datacollector.collect(self)

    def step(self):
        # tell all the agents in the model to run their step function
        self.datacollector.collect(self)
        self.schedule.step()
        self.protocol.v3.step()

    def run_model(self):
        for i in range(self.max_steps):
            print(self.protocol.v3.global_state.timestamp)
            self.step()



# coding=utf-8
# --------------------------------------------------------------------------------------------------------------------
# Licensed under the MIT LICENSE. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------------------------------
"""System state variables, constants, containers and interface."""
import copy, logging, pandas
from dataclasses import field
from pydantic.types import Tuple, Any, List, Dict
from pydantic.dataclasses import dataclass
from pydantic.schema import defaultdict

from bancor_research import DEFAULT, Decimal, PandasDataFrame

logger = logging.getLogger(__name__)


# Configurable Genesis Variables
DEFAULT_TIMESTAMP = DEFAULT.TIMESTAMP
DEFAULT_USERS = DEFAULT.USERS
DEFAULT_DECIMALS = DEFAULT.DECIMALS
DEFAULT_WITHDRAWAL_FEE = Decimal(DEFAULT.WITHDRAWAL_FEE[:-1]) / 100
DEFAULT_TRADING_FEE = Decimal(DEFAULT.TRADING_FEE[:-1]) / 100
DEFAULT_NETWORK_FEE = Decimal(DEFAULT.NETWORK_FEE[:-1]) / 100
DEFAULT_BNT_FUNDING_LIMIT = Decimal(DEFAULT.BNT_FUNDING_LIMIT)
DEFAULT_BNT_MIN_LIQUIDITY = Decimal(DEFAULT.BNT_MIN_LIQUIDITY)
DEFAULT_COOLDOWN_TIME = DEFAULT.COOLDOWN_TIME
DEFAULT_ALPHA = Decimal("0.2")
DEFAULT_LOWER_EMA_LIMIT = Decimal("0.99")
DEFAULT_UPPER_EMA_LIMIT = Decimal("1.01")
DEFAULT_ACCOUNT_BALANCE = Decimal("0")


# Misc dependencies for `State`
class Token(object):
    """
    Represents a token balance with common math operations to increase, decrease, and set the balance.
    """

    def __init__(self, balance: Decimal = Decimal("0")):
        self.balance = balance

    def add(self, value: Decimal):
        self.validate_balance()
        self.balance += self.validate(value)

    def subtract(self, value: Decimal):
        self.validate_balance()
        self.balance -= self.validate(value)

    def set(self, value: Decimal):
        self.validate_balance()
        self.balance = self.validate(value)

    def validate(self, value) -> Decimal:
        return self.validate_value(value)

    def validate_balance(self):
        if pandas.isnull(self.balance):
            self.balance = Decimal("0")

    def validate_value(self, value) -> Decimal:
        if pandas.isnull(value):
            value = Decimal("0")
        return Decimal(str(value))


# Containers
class Config:
    validate_assignment = True
    arbitrary_types_allowed = True


@dataclass(config=Config)
class GlobalSettings:
    """
    Represents the default global settings. These can be overridden by the BancorDapp configuration upon instantiation.
    """

    timestamp: int = DEFAULT_TIMESTAMP
    whitelisted_tokens: dict = field(default_factory=dict)
    active_users: List[str] = field(default_factory=lambda: DEFAULT_USERS)
    cooldown_time: int = DEFAULT_COOLDOWN_TIME
    network_fee: Decimal = DEFAULT_NETWORK_FEE
    withdrawal_fee: Decimal = DEFAULT_WITHDRAWAL_FEE
    bnt_min_liquidity: Decimal = DEFAULT_BNT_MIN_LIQUIDITY
    alpha: Decimal = DEFAULT_ALPHA


@dataclass(config=Config)
class Cooldown:
    """
    Represents a pending withdrawal cooldown.
    """

    id: int
    created_at: int
    user_name: str
    tkn_name: str
    is_complete: bool
    tkn: Any = field(default_factory=Token)
    pooltoken: Any = field(default_factory=Token)


@dataclass(config=Config)
class StandardProgram:
    """
    Represents an standard reward program state.
    """

    id: int
    tkn_name: str
    is_active: bool
    start_time: int
    end_time: int
    last_update_time: int
    reward_rate: Decimal
    remaining_rewards: Any = field(default_factory=Token)
    reward_per_token: Any = field(default_factory=Token)
    total_unclaimed_rewards: Any = field(default_factory=Token)
    staked_reward_amt: Any = field(default_factory=Token)
    pooltoken_amt: Any = field(default_factory=Token)


@dataclass(config=Config)
class AutocompoundingProgram:
    """
    Represents an autocompounding reward program state.
    """

    id: int
    created_at: int
    tkn_name: str
    distribution_type: str
    start_time: int
    total_duration: int = 0
    half_life: int = 0
    total_rewards: Any = field(default_factory=Token)
    remaining_rewards: Any = field(default_factory=Token)
    prev_token_amt_distributed: Any = field(default_factory=Token)
    is_active: bool = False
    is_enabled: bool = False

    @property
    def flat_distribution_rate_per_second(self):
        """
        Returns the rate per second of the distribution.
        """
        return self.total_rewards.balance / self.total_duration


@dataclass(config=Config)
class UserStandardProgram:
    """
    Represents a standard reward program user state
    """

    staked_amt: Any = field(default_factory=Token)
    pending_rewards: Any = field(default_factory=Token)
    reward_per_token_paid: Any = field(default_factory=Token)


@dataclass(config=Config)
class User:
    """
    Represents a user agent state.
    """

    user_name: str
    pending_withdrawals: Dict[int, Cooldown] = field(
        default_factory=lambda: defaultdict(Cooldown)
    )
    pending_standard_rewards: Dict[int, UserStandardProgram] = field(
        default_factory=lambda: defaultdict(UserStandardProgram)
    )
    wallet: Dict[str, Token] = field(default_factory=lambda: defaultdict(Token))


@dataclass(config=Config)
class Tokens(GlobalSettings):
    """
    Represents all ledger and other configuration balances associated with a particular token's current state.
    """

    tkn_name: str = None
    decimals: int = DEFAULT_DECIMALS
    master_vault: Any = field(default_factory=Token)
    staking_ledger: Any = field(default_factory=Token)
    pooltoken_supply: Any = field(default_factory=Token)
    protocol_wallet_pooltokens: Any = field(default_factory=Token)
    vortex_ledger: Any = field(default_factory=Token)
    vbnt_burned: Any = field(default_factory=Token)
    external_protection_vault: Any = field(default_factory=Token)
    external_rewards_vault: Any = field(default_factory=Token)
    bnt_trading_liquidity: Any = field(default_factory=Token)
    tkn_trading_liquidity: Any = field(default_factory=Token)
    bnt_funding_amt: Any = field(default_factory=Token)
    spot_rate: Decimal = Decimal("0")
    ema_rate: Decimal = Decimal("0")
    inv_spot_rate: Decimal = Decimal("0")
    inv_ema_rate: Decimal = Decimal("0")
    ema_last_updated: Decimal = Decimal("0")
    is_trading_enabled: bool = False
    bnt_funding_limit: Decimal = DEFAULT_BNT_FUNDING_LIMIT
    trading_fee: Decimal = DEFAULT_TRADING_FEE

    @property
    def bnt_remaining_funding(self):
        """
        Computes the BNT funding remaining for the pool.
        """
        return self.bnt_funding_limit - self.bnt_funding_amt.balance

    @property
    def is_price_stable(self):
        """
        True if the spot price deviation from the EMA is less than 1% (or other preset threshold amount).
        """
        # changed due to contract implementation
        # should probably use `ema_rate` and `inv_ema_rate` instead
        ema_rate = self.updated_ema_rate
        inv_ema_rate = self.updated_inv_ema_rate
        return (
            DEFAULT_LOWER_EMA_LIMIT * ema_rate
            <= self.spot_rate
            <= DEFAULT_UPPER_EMA_LIMIT * ema_rate
            and DEFAULT_LOWER_EMA_LIMIT * inv_ema_rate
            <= self.inv_spot_rate
            <= DEFAULT_UPPER_EMA_LIMIT * inv_ema_rate
        )

    @property
    def updated_ema_rate(self) -> Decimal:
        """
        Computes the ema as a lagging average only once per block, per pool.
        """
        return self.alpha * self.spot_rate + (1 - self.alpha) * self.ema_rate

    @property
    def updated_inv_ema_rate(self) -> Decimal:
        """
        Computes the inverse ema as a lagging average only once per block, per pool.
        """
        return self.alpha * self.inv_spot_rate + (1 - self.alpha) * self.inv_ema_rate

    @property
    def bnt_bootstrap_liquidity(self):
        """
        Computes the minimum between bnt_min_liquidity multiplied by 2 and bnt_funding_limit.
        """
        return min(2 * self.bnt_min_liquidity, self.bnt_funding_limit)

    @property
    def vbnt_price(self):
        """
        Returns the price of the current vbnt token. Only valid when name==bnt
        """
        assert (
            self.tkn_name == "bnt"
        ), f"vbnt_price attempted to be accessed in {self.tkn_name} state, call bnt state instead"
        return self._vbnt_price


@dataclass(config=Config)
class State(GlobalSettings):
    """
    Represents the system state at the current timestamp. Main interface for all other dataclasses.
    """

    transaction_id: int = 0
    price_feeds: PandasDataFrame = None
    tokens: Dict[str, Tokens] = field(default_factory=lambda: defaultdict(Tokens))
    users: Dict[str, User] = field(default_factory=lambda: defaultdict(User))
    standard_reward_programs: Dict[int, StandardProgram] = field(
        default_factory=lambda: defaultdict(StandardProgram)
    )
    autocompounding_reward_programs: Dict[str, AutocompoundingProgram] = field(
        default_factory=lambda: defaultdict(AutocompoundingProgram)
    )
    history: list = field(default_factory=list)
    logger: Any = logger
    json_export: dict = field(default_factory=dict)
    whitelisted_tokens: dict = field(default_factory=dict)
    rolling_trade_fees: dict = field(default_factory=dict)

    @property
    def valid_rewards_programs(self):
        """
        Returns all valid autocompounding programs for the current state.
        """
        return [
            p
            for p in self.autocompounding_reward_programs
            if self.autocompounding_reward_programs[p].is_active
            and self.autocompounding_reward_programs[p].is_enabled
        ]

    @property
    def usernames(self):
        """
        Returns a list of all current users
        """
        return [user for user in self.users]

    @property
    def withdrawal_ids(self):
        """
        Returns a list of all withdrawal_ids
        """
        return sum([len(self.users[user].pending_withdrawals) for user in self.users])

    @property
    def autocompounding_programs_count(self) -> int:
        """
        Returns the count of active autocompounding reward programs
        """
        return len(
            [
                self.autocompounding_reward_programs[tkn_name]
                for tkn_name in self.autocompounding_reward_programs
                if self.autocompounding_reward_programs[tkn_name].is_active
            ]
        )

    @property
    def active_autocompounding_programs(self) -> list:
        """
        Returns the active autocompounding reward programs
        """
        return [
            tkn_name
            for tkn_name in self.autocompounding_reward_programs
            if self.autocompounding_reward_programs[tkn_name].is_active
        ]

    @property
    def active_standard_programs(self) -> list:
        """
        Returns the active standard reward programs
        """
        return [
            tkn_name
            for tkn_name in self.standard_reward_programs
            if self.standard_reward_programs[tkn_name].is_active
        ]

    @property
    def standard_programs_count(self) -> int:
        """
        Returns the count of active standard reward programs
        """
        return len(self.active_standard_programs())

    @property
    def bnt_price(self) -> Decimal:
        """
        Returns the bnt price feed at the current timestamp.
        """
        return Decimal(self.price_feeds.at[self.timestamp, "bnt"])

    @property
    def bnt_virtual_balance(self) -> Decimal:
        """
        Returns the inverse of the bnt price feed at the current timestamp
        """
        return Decimal("1") / self.bnt_price

    @property
    def bnbnt_rate(self) -> Decimal:
        """
        Returns the inverse of the bnt price feed at the current timestamp
        """
        pooltoken_supply = self.tokens["bnt"].pooltoken_supply.balance
        staking_ledger = self.tokens["bnt"].staking_ledger.balance
        if pooltoken_supply != staking_ledger:
            return Decimal(pooltoken_supply) / Decimal(staking_ledger)
        return Decimal("1")

    def step(self):
        """
        Incriments the current timestamp.
        """
        self.timestamp += 1

    def set_external_rewards_vault_balance(self, tkn_name: str, value: Decimal):
        """
        Set external rewards vault balance by a given amount.
        """
        self.tokens[tkn_name].external_rewards_vault.set(value)

    def map_user_standard_program(self, user_name: str, id: int):
        """
        Map a user standard program if needed.
        """
        if id not in self.users[user_name].pending_standard_rewards:
            self.users[user_name].pending_standard_rewards[id] = UserStandardProgram()

    def set_user_pending_standard_rewards(
        self, user_name: str, id: int, value: Decimal
    ):
        """
        Set standard rewards vault balance by a given amount.
        """
        self.users[user_name].pending_standard_rewards[id].pending_rewards.set(value)

    def increase_external_rewards_vault_balance(self, tkn_name: str, value: Decimal):
        """
        Increase external rewards vault balance by a given amount.
        """
        self.tokens[tkn_name].external_rewards_vault.add(value)

    def decrease_external_rewards_vault_balance(self, tkn_name: str, value: Decimal):
        """
        Decrease external rewards vault balance by a given amount.
        """
        self.tokens[tkn_name].external_rewards_vault.subtract(value)

    def decrease_pooltoken_balance(self, tkn_name: str, value: Decimal):
        """
        Decrease staked_amt supply balance by a given amount.
        """
        self.tokens[tkn_name].pooltoken_supply.subtract(value)

    def increase_pooltoken_balance(self, tkn_name: str, value: Decimal):
        """
        Increase staked_amt supply balance by a given amount.
        """
        self.tokens[tkn_name].pooltoken_supply.add(value)

    def increase_user_standard_rewards_stakes(
        self, id: int, user_name: str, value: Decimal
    ):
        """
        Increase user standard rewards staked_reward_amt pooltokens by a given amount.
        """
        self.users[user_name].pending_standard_rewards[id].staked_amt.add(value)
        self.standard_reward_programs[id].staked_reward_amt.add(value)

    def decrease_user_standard_rewards_stakes(
        self, id: int, user_name: str, value: Decimal
    ):
        """
        Increase user standard rewards staked_reward_amt pooltokens by a given amount.
        """
        self.users[user_name].pending_standard_rewards[id].staked_amt.subtract(value)
        self.standard_reward_programs[id].staked_reward_amt.subtract(value)

    def set_user_standard_rewards_stakes(self, id: int, user_name: str, value: Decimal):
        """
        Set user standard rewards staked_reward_amt pooltokens to a given amount.
        """
        self.users[user_name].pending_standard_rewards[id].staked_amt.set(value)

    def decrease_bnt_trading_liquidity(self, tkn_name: str, value: Decimal):
        """
        Decrease bnt_trading_liquidity balance by a given amount.
        """
        self.tokens[tkn_name].bnt_trading_liquidity.subtract(value)

    def increase_bnt_trading_liquidity(self, tkn_name: str, value: Decimal):
        """
        Increase bnt_trading_liquidity balance by a given amount.
        """
        self.tokens[tkn_name].bnt_trading_liquidity.add(value)

    def decrease_tkn_trading_liquidity(self, tkn_name: str, value: Decimal):
        """
        Decrease tkn_trading_liquidity balance by a given amount.
        """
        self.tokens[tkn_name].tkn_trading_liquidity.subtract(value)

    def increase_tkn_trading_liquidity(self, tkn_name: str, value: Decimal):
        """
        Increase tkn_trading_liquidity balance by a given amount.
        """
        self.tokens[tkn_name].tkn_trading_liquidity.add(value)

    def decrease_bnt_funding_amt(self, tkn_name: str, value: Decimal):
        """
        Decrease bnt_funding_amt balance by a given amount.
        """
        self.tokens[tkn_name].bnt_funding_amt.subtract(value)

    def increase_bnt_funding_amt(self, tkn_name: str, value: Decimal):
        """
        Increase bnt_funding_amt balance by a given amount.
        """
        self.tokens[tkn_name].bnt_funding_amt.add(value)

    def decrease_protocol_wallet_balance(self, tkn_name: str, value: Decimal):
        """
        Decrease protocol wallet balance by a given amount.
        """
        self.tokens[tkn_name].protocol_wallet_pooltokens.subtract(value)

    def increase_protocol_wallet_balance(self, tkn_name: str, value: Decimal):
        """
        Increase protocol wallet balance by a given amount.
        """
        self.tokens[tkn_name].protocol_wallet_pooltokens.add(value)

    def decrease_master_vault_balance(self, tkn_name: str, value: Decimal):
        """
        Decrease master_vault balance by a given amount.
        """
        self.tokens[tkn_name].master_vault.subtract(value)

    def increase_master_vault_balance(self, tkn_name: str, value: Decimal):
        """
        Increase master_vault balance by a given amount.
        """
        self.tokens[tkn_name].master_vault.add(value)

    def decrease_staked_balance(self, tkn_name: str, value: Decimal):
        """
        Decrease staked_reward_amt balance by a given amount.
        """
        self.tokens[tkn_name].staking_ledger.subtract(value)

    def increase_staked_balance(self, tkn_name: str, value: Decimal):
        """
        Increase staked_reward_amt balance by a given amount.
        """
        self.tokens[tkn_name].staking_ledger.add(value)

    def decrease_external_protection_balance(self, tkn_name: str, value: Decimal):
        """
        Decrease external protection balance by a given amount.
        """
        self.tokens[tkn_name].external_protection_vault.subtract(value)

    def increase_external_protection_balance(self, tkn_name: str, value: Decimal):
        """
        Increase external protection balance by a given amount.
        """
        self.tokens[tkn_name].external_protection_vault.add(value)

    def decrease_user_balance(self, user_name: str, tkn_name: str, value: Decimal):
        """
        Decrease user balance by a given amount.
        """
        self.users[user_name].wallet[tkn_name].subtract(value)

    def increase_user_balance(self, user_name: str, tkn_name: str, value: Decimal):
        """
        Increase user balance by a given amount.
        """
        self.users[user_name].wallet[tkn_name].add(value)

    def set_vortex_balance(self, tkn_name: str, value: Decimal):
        """
        Set vortex_ledger balance to a given amount.
        """
        self.tokens[tkn_name].vortex_ledger.set(value)

    def decrease_vbnt_burned(self, tkn_name: str, value: Decimal):
        """
        Decrease vbnt amount burned.
        """
        self.tokens[tkn_name].vbnt_burned.subtract(value)

    def increase_vbnt_burned(self, tkn_name: str, value: Decimal):
        """
        Increase vbnt amount burned.
        """
        self.tokens[tkn_name].vbnt_burned.add(value)

    def set_vbnt_burned(self, tkn_name: str, value: Decimal):
        """
        Set vbnt amount burned.
        """
        self.tokens[tkn_name].vbnt_burned.set(value)

    def increase_vortex_balance(self, tkn_name: str, value: Decimal):
        """
        Increase vortex_ledger balance by a given amount.
        """
        self.tokens[tkn_name].vortex_ledger.add(value)

    def decrease_vortex_balance(self, tkn_name: str, value: Decimal):
        """
        Decrease vortex_ledger balance by a given amount.
        """
        self.tokens[tkn_name].vortex_ledger.subtract(value)

    def set_standard_rewards_per_token(self, id: int, value: Decimal):
        """
        Set the standard reward program rewards per token.
        """
        self.standard_reward_programs[id].reward_per_token.set(value)

    def set_standard_rewards_last_update_time(self, id: int, value: int):
        """
        Set the standard reward program last updated time.
        """
        self.standard_reward_programs[id].last_update_time = value

    def set_provider_reward_per_token_paid(
        self, user_name: str, id: int, value: Decimal
    ):
        """
        Set the user's rewards per token paid for a given program.
        """
        self.users[user_name].pending_standard_rewards[id].reward_per_token_paid.set(
            value
        )

    def set_provider_pending_standard_rewards(
        self, user_name: str, id: int, value: Decimal
    ):
        """
        Set the user's pending standard rewards for a given program.
        """
        self.users[user_name].pending_standard_rewards[id].pending_rewards.set(value)

    def set_trading_fee(self, tkn_name: str, value: Decimal):
        """
        Set the trading fee for a given tkn_name.
        """
        self.tokens[tkn_name].trading_fee = value

    def set_bnt_funding_limit(self, tkn_name: str, value: Decimal):
        """
        Set the bnt funding limit for a given tkn_name.
        """
        self.tokens[tkn_name].bnt_funding_limit = value

    def set_bnt_trading_liquidity(self, tkn_name: str, value: Decimal):
        """
        Set the bnt_trading_liquidity balance to a given amount.
        """
        self.tokens[tkn_name].bnt_trading_liquidity.set(value)

    def set_standard_program_end_time(self, id: int, value: int):
        """
        Set the end time of the standard program.
        """
        self.standard_reward_programs[id].end_time = value

    def set_standard_program_is_active(self, id: int, value: bool):
        """
        Set the is_active status of a standard program.
        """
        self.standard_reward_programs[id].is_active = value

    def set_tkn_trading_liquidity(self, tkn_name: str, value: Decimal):
        """
        Set the tkn_trading_liquidity balance to a given amount.
        """
        self.tokens[tkn_name].tkn_trading_liquidity.set(value)

    def set_withdrawal_fee(self, tkn_name: str, value: Decimal):
        """
        Set the withdrawal_fee to a given amount.
        """
        self.tokens[tkn_name].withdrawal_fee = value

    def set_prev_token_amt_distributed(self, tkn_name: str, value: Decimal):
        """
        Set the prev_token_amt_distributed to a given value.
        """
        self.autocompounding_reward_programs[tkn_name].prev_token_amt_distributed.set(
            value
        )

    def set_is_trading_enabled(self, tkn_name: str, value: bool):
        """
        Set the set_is_trading_enabled to a given status.
        """
        self.tokens[tkn_name].is_trading_enabled = value

    def set_spot_rate(self, tkn_name: str, value: Decimal):
        """
        Set the spot rate to a given value.
        """
        self.tokens[tkn_name].spot_rate = value

    def set_ema_rate(self, tkn_name: str, value: Decimal):
        """
        Set the ema rate to a given amount.
        """
        self.tokens[tkn_name].ema_rate = value

    def set_inv_spot_rate(self, tkn_name: str, value: Decimal):
        """
        Set the inverse spot rate to a given value.
        """
        self.tokens[tkn_name].inv_spot_rate = value

    def set_inv_ema_rate(self, tkn_name: str, value: Decimal):
        """
        Set the inverse ema rate to a given value.
        """
        self.tokens[tkn_name].inv_ema_rate = value

    def set_bnt_funding_amt(self, tkn_name: str, value: Decimal):
        """
        Set the bnt_funding_amt to a given status.
        """
        self.tokens[tkn_name].bnt_funding_amt.set(value)

    def set_initial_rates(self, tkn_name: str, bootstrap_rate: Decimal):
        self.tokens[tkn_name].spot_rate = bootstrap_rate
        self.tokens[tkn_name].ema_rate = bootstrap_rate
        self.tokens[tkn_name].inv_spot_rate = bootstrap_rate**-1
        self.tokens[tkn_name].inv_ema_rate = bootstrap_rate**-1

    def set_staked_balance(self, tkn_name: str, value: Decimal):
        """
        Set the staked_reward_amt balance to a given amount.
        """
        self.tokens[tkn_name].staking_ledger.set(value)

    def set_protocol_wallet_balance(self, tkn_name: str, value: Decimal):
        """
        Set the protocol_wallet balance to a given amount.
        """
        self.tokens[tkn_name].protocol_wallet_pooltokens.set(value)

    def set_user_balance(self, user_name: str, tkn_name: str, value: Decimal):
        """
        Set the user balance to a given amount.
        """
        self.users[user_name].wallet[tkn_name].set(value)

    def set_network_fee(self, tkn_name: str, value: Decimal):
        """
        Set the network fee to a given amount.
        """
        self.tokens[tkn_name].network_fee = value

    def set_pooltoken_balance(self, tkn_name: str, value: Decimal):
        """
        Set the staked_reward_amt balance to a given amount.
        """
        self.tokens[tkn_name].pooltoken_supply.set(value)

    def set_standard_remaining_rewards(self, id: int, value: Decimal):
        """
        Set the remaining_rewards balance to a given amount.
        """
        self.standard_reward_programs[id].remaining_rewards.set(value)

    def set_autocompounding_remaining_rewards(self, tkn_name: str, value: Decimal):
        """
        Set the remaining_rewards balance to a given amount.
        """
        self.autocompounding_reward_programs[tkn_name].remaining_rewards.set(value)

    def set_token_amt_to_distribute(self, tkn_name: str, value: Decimal):
        """
        Set the token_amt_to_distribute balance to a given amount.
        """
        self.autocompounding_reward_programs[tkn_name].token_amt_to_distribute.set(
            value
        )

    def set_program_is_active(self, tkn_name: str, value: bool):
        """
        Set the rewards is_active flag to a given status.
        """
        self.autocompounding_reward_programs[tkn_name].is_active = value

    def set_pending_withdrawals_status(
        self, user_name: str, id_number: int, status: bool
    ):
        """
        Set pending_withdrawal to a given status.
        """
        self.users[user_name].pending_withdrawals[id_number].is_complete = status

    def create_whitelisted_tkn(self, tkn_name: str):
        """
        Adds a new tkn_name to the whitelisted_tokens
        """
        if tkn_name not in self.whitelisted_tokens:
            self.whitelisted_tokens[tkn_name] = {
                "decimals": 18,
                "trading_fee": DEFAULT_TRADING_FEE,
                "bnt_funding_limit": DEFAULT_BNT_FUNDING_LIMIT,
                "ep_vault_balance": DEFAULT.EP_VAULT_BALANCE,
            }
        return self

    def create_user(self, user_name: str):
        """
        Creates a new system user agent.
        """
        if user_name not in self.users:
            self.users[user_name] = User(user_name=user_name)

        for tkn_name in self.whitelisted_tokens:
            if tkn_name not in self.users[user_name].wallet:
                self.users[user_name].wallet[tkn_name] = Token(
                    balance=DEFAULT_ACCOUNT_BALANCE
                )
            if get_pooltoken_name(tkn_name) not in self.users[user_name].wallet:
                self.users[user_name].wallet[get_pooltoken_name(tkn_name)] = Token(
                    balance=DEFAULT_ACCOUNT_BALANCE
                )
        if "bnt" not in self.users[user_name].wallet:
            self.users[user_name].wallet["bnt"] = Token(balance=DEFAULT_ACCOUNT_BALANCE)
        if "vbnt" not in self.users[user_name].wallet:
            self.users[user_name].wallet["vbnt"] = Token(
                balance=DEFAULT_ACCOUNT_BALANCE
            )
        return self

    def update_spot_rate(self, tkn_name: str):
        """
        Updates the spot rate and the inverse spot rate.
        """
        bnt_trading_liquidity = get_bnt_trading_liquidity(self, tkn_name)
        tkn_trading_liquidity = get_tkn_trading_liquidity(self, tkn_name)
        if bnt_trading_liquidity == 0 and tkn_trading_liquidity == 0:
            spot_rate = Decimal(0)
            inv_spot_rate = Decimal(0)
        else:
            spot_rate = bnt_trading_liquidity / tkn_trading_liquidity
            inv_spot_rate = tkn_trading_liquidity / bnt_trading_liquidity
        self.set_spot_rate(tkn_name, spot_rate)
        self.set_inv_spot_rate(tkn_name, inv_spot_rate)

    def next_standard_program_id(self) -> int:
        """
        Returns the next standard program ID.
        """
        return len(self.standard_reward_programs) + 1

    def copy(self):
        return copy.deepcopy(self)


def get_total_standard_rewards_staked(state, id: int) -> Decimal:
    """
    Returns the total standard rewards staked_reward_amt for a given program id.
    """
    return state.standard_reward_programs[id].staked_reward_amt.balance


def get_tkn_virtual_balance(state: State, tkn_name: str) -> Decimal:
    """
    Return the reciprocal of the price of bnt or tkn.
    """
    return Decimal("1") / get_tkn_price(state, tkn_name)


def get_virtual_rate(state: State, tkn_name: str) -> Decimal:
    """
    Returns bnt_virtual_balance / tkn_token_virtual_balance
    """
    return get_bnt_virtual_balance(state) / get_tkn_virtual_balance(state, tkn_name)


def get_bnt_virtual_balance(state: State) -> Decimal:
    """
    Returns bnt_virtual_balance via the common get interface for consistency
    """
    return state.bnt_virtual_balance


def get_tkn_bootstrap_liquidity(state: State, tkn_name: str) -> Decimal:
    """
    The tkn equivalnce of twice the bnt min liquidity (i.e. the bnt bootstrap liquidity)
    """
    return state.tokens[tkn_name].bnt_bootstrap_liquidity / get_virtual_rate(
        state, tkn_name
    )


def get_bnt_bootstrap_liquidity(state: State, tkn_name: str) -> Decimal:
    """
    Returns the current state bnt bootstrap liquidity via get interface for consistency
    """
    return state.tokens[tkn_name].bnt_bootstrap_liquidity


def get_staked_balance(state: State, tkn_name: str) -> Decimal:
    """
    The current balance of the staking ledger for a given tkn_name.
    """
    return state.tokens[tkn_name].staking_ledger.balance


def get_master_vault_balance(state: State, tkn_name: str) -> Decimal:
    """
    The current balance of the master_vault for a given tkn_name.
    """
    return state.tokens[tkn_name].master_vault.balance


def get_bnt_trading_liquidity(state: State, tkn_name: str) -> Decimal:
    """
    The current bnt trading liquidity for a given tkn_name.
    """
    return state.tokens[tkn_name].bnt_trading_liquidity.balance


def get_standard_reward_pool_token_name(state: State, id: int) -> str:
    """
    Get the standard reward pool token name for a given id.
    """
    return get_pooltoken_name(state.standard_reward_programs[id].tkn_name)


def get_standard_reward_per_token(state: State, id: int) -> Decimal:
    """
    Get the standard reward per token for a given id.
    """
    return state.standard_reward_programs[id].reward_per_token.balance


def get_standard_reward_end_time(state: State, id: int) -> int:
    """
    Get the standard reward end time for a given id.
    """
    return state.standard_reward_programs[id].end_time


def get_standard_reward_start_time(state: State, id: int) -> int:
    """
    Get the standard reward start time for a given id.
    """
    return state.standard_reward_programs[id].start_time


def get_standard_reward_last_update_time(state: State, id: int) -> int:
    """
    Get the standard reward last update time for a given id.
    """
    return state.standard_reward_programs[id].last_update_time


def get_standard_reward_rate(state: State, id: int) -> int:
    """
    Get the standard reward rate for a given id.
    """
    return state.standard_reward_programs[id].reward_rate


def get_user_pending_rewards_staked_balance(
    state: State, id: int, user_name: str
) -> Decimal:
    return state.users[user_name].pending_standard_rewards[id].staked_amt.balance


def get_user_pending_standard_rewards(state: State, id: int, user_name: str) -> Decimal:
    return state.users[user_name].pending_standard_rewards[id].pending_rewards.balance


def get_user_reward_per_token_paid(state: State, id: int, user_name: str) -> Decimal:
    return (
        state.users[user_name]
        .pending_standard_rewards[id]
        .reward_per_token_paid.balance
    )


def get_tkn_trading_liquidity(state: State, tkn_name: str) -> Decimal:
    """
    The current tkn trading liquidity for a given tkn_name.
    """
    return state.tokens[tkn_name].tkn_trading_liquidity.balance


def get_bnt_funding_amt(state: State, tkn_name: str) -> Decimal:
    """
    The current bnt amount funded for a given tkn_name.
    """
    return state.tokens[tkn_name].bnt_funding_amt.balance


def get_external_protection_vault_balance(state: State, tkn_name: str) -> Decimal:
    """
    The current external protection vault balance for a given tkn_name.
    """
    return state.tokens[tkn_name].external_protection_vault.balance


def get_external_rewards_vault_balance(state: State, tkn_name: str) -> Decimal:
    """
    The current external rewards vault balance for a given tkn_name.
    """
    return state.tokens[tkn_name].external_rewards_vault.balance


def get_pooltoken_balance(state: State, tkn_name: str) -> Decimal:
    """
    The current erc20 contracts staked_amt supply for a given tkn_name.
    """
    return state.tokens[tkn_name].pooltoken_supply.balance


def get_protocol_wallet_balance(state: State, tkn_name: str) -> Decimal:
    """
    The current protocol owned liquidity (when tkn_name==bnt) or the current autocompounding rewards remaining.
    """
    return state.tokens[tkn_name].protocol_wallet_pooltokens.balance


def get_vortex_balance(state: State, tkn_name: str) -> Decimal:
    """
    The current vortex_ledger balance for a given tkn_name.
    """
    return state.tokens[tkn_name].vortex_ledger.balance


def get_user_balance(state: State, user_name: str, tkn_name: str) -> Decimal:
    """
    The user balance for a given tkn_name.
    """
    return state.users[user_name].wallet[tkn_name].balance


def get_bnbnt_rate(state: State) -> Decimal:
    """
    The current bnbnt rate.
    """
    return state.bnbnt_rate


def get_total_rewards(state: State, tkn_name: str) -> Decimal:
    """
    The initial balance for a given autocompounding rewards program.
    """
    return state.autocompounding_reward_programs[tkn_name].total_rewards.balance


def get_distribution_type(state: State, tkn_name: str) -> str:
    """
    The type of distribution for an autocompounding rewards program (either flat or exponential "exp")
    """
    return state.autocompounding_reward_programs[tkn_name].distribution_type


def get_prev_token_amt_distributed(state: State, tkn_name: str) -> Decimal:
    """
    The previous distribution amount for a given autocompounding rewards program.
    """
    return state.autocompounding_reward_programs[
        tkn_name
    ].prev_token_amt_distributed.balance


def get_autocompounding_start_time(state: State, tkn_name: str) -> int:
    """
    The start time for a given autocompounding rewards program.
    """
    return state.autocompounding_reward_programs[tkn_name].start_time


def get_half_life(state: State, tkn_name: str) -> int:
    """
    The half life for a given autocompounding rewards program.
    """
    return state.autocompounding_reward_programs[tkn_name].half_life


def get_timestamp(state: State) -> int:
    """
    The current state timestamp.
    """
    return state.timestamp


def get_avg_tkn_trading_liquidity(
    state: State, tkn_name: str, rate: Decimal
) -> Decimal:
    """
    The average trading liquidity for a given tkn_name.
    """
    return get_bnt_trading_liquidity(state, tkn_name) / rate


def get_updated_ema_rate(state: State, tkn_name: str) -> Decimal:
    """
    The ema as a lagging average only once per block, per pool.
    """
    if state.tokens[tkn_name].ema_last_updated == state.tokens[tkn_name].timestamp:
        return state.tokens[tkn_name].ema_rate
    return state.tokens[tkn_name].updated_ema_rate


def get_updated_inv_ema_rate(state: State, tkn_name: str) -> Decimal:
    """
    The inverse ema as a lagging average only once per block, per pool.
    """
    if state.tokens[tkn_name].ema_last_updated == state.tokens[tkn_name].timestamp:
        return state.tokens[tkn_name].inv_ema_rate
    return state.tokens[tkn_name].updated_inv_ema_rate


def get_tkn_excess_bnt_equivalence(
    state: State, tkn_name: str, rate: Decimal
) -> Decimal:
    """
    Returns the equivalent bnt value of the non-trading tkn balance of the master_vault.
    """
    return get_tkn_excess(state, tkn_name) * rate


def get_tkn_excess(state: State, tkn_name: str) -> Decimal:
    """
    The difference between the master_vault balance and the average trading liquidity.
    """
    return get_master_vault_balance(state, tkn_name) - get_tkn_trading_liquidity(
        state, tkn_name
    )


def get_is_trading_enabled(state: State, tkn_name: str) -> bool:
    """
    Returns True if trading is enabled for a given tkn_name.
    """
    return state.tokens[tkn_name].is_trading_enabled


def get_is_price_stable(state: State, tkn_name: str) -> bool:
    """
    Returns True if the spot_price vs EMA deviation is less than 1%.
    """
    return state.tokens[tkn_name].is_price_stable


def get_bnt_remaining_funding(state: State, tkn_name: str) -> Decimal:
    """
    Returns the bnt funding remaining for a given tkn_name relative to the limit.
    """
    return state.tokens[tkn_name].bnt_remaining_funding


def get_flat_distribution_rate_per_second(state: State, tkn_name: str) -> Decimal:
    """
    Returns the flat distribution rate per second for a given tkn_name based on the program duration.
    """
    return state.autocompounding_reward_programs[
        tkn_name
    ].flat_distribution_rate_per_second


def get_spot_rate(state: State, tkn_name: str) -> Decimal:
    """
    Returns the spot rate for a given tkn_name.
    """
    return state.tokens[tkn_name].spot_rate


def get_ema_rate(state: State, tkn_name: str) -> Decimal:
    """
    Returns the ema rate for a given tkn_name.
    """
    return state.tokens[tkn_name].ema_rate


def get_inv_spot_rate(state: State, tkn_name: str) -> Decimal:
    """
    Returns the inverse spot rate for a given tkn_name.
    """
    return state.tokens[tkn_name].inv_spot_rate


def get_inv_ema_rate(state: State, tkn_name: str) -> Decimal:
    """
    Returns the inverse ema rate for a given tkn_name.
    """
    return state.tokens[tkn_name].inv_ema_rate


def get_ema_last_updated(state: State, tkn_name: str) -> Decimal:
    """
    Returns the timestamp of the last EMA update of a given tkn_name.
    """
    return state.tokens[tkn_name].ema_last_updated


def get_whitelisted_tokens(state: State) -> List[str]:
    """
    Returns the current state whitelisted tokens.
    """
    return state.whitelisted_tokens


def get_usernames(state: State) -> List[str]:
    """
    Returns the current state users by usernames.
    """
    return state.usernames


def get_trading_fee(state: State, tkn_name: str) -> Decimal:
    """
    Returns the trading fee for a given tkn_name.
    """
    return state.tokens[tkn_name].trading_fee


def get_bnt_funding_limit(state: State, tkn_name: str) -> Decimal:
    """
    Returns the bnt funding limit for a given tkn_name.
    """
    return state.tokens[tkn_name].bnt_funding_limit


def get_withdrawal_id(state: State) -> int:
    """
    Generate withdraw id.
    """
    return state.withdrawal_ids


def get_tkn_price(state: State, tkn_name: str) -> Decimal:
    """
    Gets the tkn price from the price feed.
    """
    return (
        state.tokens[tkn_name].vbnt_price
        if tkn_name == "vbnt"
        else Decimal(state.price_feeds.at[state.timestamp, tkn_name])
    )


def get_user_pending_withdrawals(state: State, user_name: str, tkn_name: str) -> list:
    """
    Returns a given user's pending withdrawals for a given tkn_name
    """
    return [
        id
        for id in state.users[user_name].pending_withdrawals
        if state.users[user_name].pending_withdrawals[id].tkn_name == tkn_name
        and state.users[user_name].pending_withdrawals[id].is_complete
    ]


def get_prices(state: State, tkn_name: str) -> Tuple[Decimal, Decimal]:
    """
    Returns tkn & bnt price from the price feed.
    """
    return get_tkn_price(state, tkn_name), get_tkn_price(state, "bnt")


def get_is_ema_update_allowed(state: State, tkn_name: str) -> bool:
    """
    Returns the is_ema_update_allowed state for a given tkn_name.
    """
    return state.tokens[tkn_name].is_ema_update_allowed


def get_total_bnt_trading_liquidity(state: State) -> Decimal:
    """
    Returns the total sum of bnt_trading_liquidity across all whitelisted tokens
    """
    return sum(
        [
            state.tokens[name].bnt_trading_liquidity.balance
            for name in state.whitelisted_tokens
            if name != "bnt"
        ]
    )


def get_bnt_min_liquidity(state: State, tkn_name: str) -> Decimal:
    """
    Returns the bnt_min_liquidity state for a given tkn_name.
    """
    return state.tokens[tkn_name].bnt_min_liquidity


def get_rate_report(state: State, tkn_name: str, qdecimals: Decimal) -> list:
    """
    Builds a structured list for current state information display.
    """
    return [
        f"Spot Rate={state.tokens[tkn_name].spot_rate}"
        f"EMA Rate={state.tokens[tkn_name].ema_rate}"
        f"Inverse Spot Rate={state.tokens[tkn_name].inv_spot_rate}"
        f"Inverse EMA Rate={state.tokens[tkn_name].inv_ema_rate}"
    ]


def get_trading_liquidity_description(state: State, qdecimals: Decimal) -> list:
    """
    Builds a structured list for current state information display.
    """
    return [
        f"bnt={get_bnt_trading_liquidity(state, tkn_name)} {tkn_name}="
        + str(get_tkn_trading_liquidity(state, tkn_name))
        for tkn_name in state.whitelisted_tokens
        if tkn_name != "bnt"
    ]


def get_master_vault_description(state: State, qdecimals: Decimal) -> list:
    """
    Builds a structured list for current state information display.
    """
    return [f"bnt={get_master_vault_balance(state, 'bnt')}"] + [
        f"{tkn_name}=" + str(get_master_vault_balance(state, tkn_name))
        for tkn_name in state.whitelisted_tokens
        if tkn_name != "bnt"
    ]


def get_staking_description(state: State, qdecimals: Decimal) -> list:
    """
    Builds a structured list for current state information display.
    """
    return [f"bnt={get_staked_balance(state, 'bnt')}"] + [
        f"{tkn_name}=" + str(get_staked_balance(state, tkn_name))
        for tkn_name in state.whitelisted_tokens
        if tkn_name != "bnt"
    ]


def get_pooltoken_description(state: State, qdecimals: Decimal) -> list:
    """
    Builds a structured list for current state information display.
    """
    return [f"bnbnt={get_pooltoken_balance(state, 'bnt')}"] + [
        f"bn{tkn_name}=" + str(get_pooltoken_balance(state, tkn_name))
        for tkn_name in state.whitelisted_tokens
        if tkn_name != "bnt"
    ]


def get_vortex_description(state: State, qdecimals: Decimal) -> list:
    """
    Builds a structured list for current state information display.
    """
    return ["bnt=" + str(get_vortex_balance(state, "bnt"))] + [
        "" for x in range(len(state.whitelisted_tokens[:-1]))
    ]


def get_external_protection_description(state: State, qdecimals: Decimal) -> list:
    """
    Builds a structured list for current state information display.
    """
    return [
        f"{tkn_name}=" + str(get_external_protection_vault_balance(state, tkn_name))
        for tkn_name in state.whitelisted_tokens
    ]


def get_protocol_wallet_description(state: State, qdecimals: Decimal) -> list:
    """
    Builds a structured list for current state information display.
    """
    return [f"bnbnt=" + str(get_protocol_wallet_balance(state, "bnt"))] + [
        "" for tkn_name in state.whitelisted_tokens[:-1]
    ]


def get_description(state: State, qdecimals: Decimal) -> dict:
    """
    Builds a dictionary to be used as a description for current state information display.
    """
    return {
        "Trading Liquidity": get_trading_liquidity_description(state, qdecimals),
        "Vault": get_master_vault_description(state, qdecimals),
        "Staking": get_staking_description(state, qdecimals),
        "ERC20 Contracts": get_pooltoken_description(state, qdecimals),
        "Vortex": get_vortex_description(state, qdecimals),
        "External Protection": get_external_protection_description(state, qdecimals),
        "Protocol WalletState": get_protocol_wallet_description(state, qdecimals),
    }


def get_total_holdings(state: State, user_name: str) -> Decimal:
    """
    Returns the total number of holdings for a given wallet name.
    """
    return state.users[user_name].total_holdings(state)


def get_pooltoken_name(tkn_name: str) -> str:
    """
    Returns the pool token name for a given token name.
    """
    return f"bn{tkn_name}"


def get_json_virtual_balances(state: State, tkn_name: str) -> dict:
    """
    Return the virtual balances for a given token name for a json test scenario.
    """
    return {
        "bntVirtualBalance": state.price_feeds[tkn_name].values[0],
        "baseTokenVirtualBalance": state.price_feeds["bnt"].values[0],
    }


def get_max_bnt_deposit(
    state: State,
    user_bnt: Decimal,
) -> Decimal:
    """
    Used in simulation only.
    """
    return max(get_pooltoken_balance(state, "bnt"), user_bnt)


def get_network_fee(state: State, tkn_name: str) -> Decimal:
    """
    Gets the network fee for a given tkn_name.
    """
    return state.tokens[tkn_name].network_fee


def get_trade_inputs(
    state: State, tkn_name: str
) -> Tuple[str, Decimal, Decimal, Decimal, Decimal]:
    """
    Gets all input data required to process trade action.
    """
    return (
        tkn_name,
        get_bnt_trading_liquidity(state, tkn_name),
        get_tkn_trading_liquidity(state, tkn_name),
        get_trading_fee(state, tkn_name),
        get_network_fee(state, tkn_name),
    )


def get_autocompounding_remaining_rewards(state: State, tkn_name: str) -> Decimal:
    """
    Get the remaining rewards for a given auto-compounding program.
    """
    return state.autocompounding_reward_programs[tkn_name].remaining_rewards.balance


def get_standard_remaining_rewards(state: State, id: int) -> Decimal:
    """
    Get the remaining rewards for a given standard program.
    """
    return state.standard_reward_programs[id].remaining_rewards.balance



# coding=utf-8
# --------------------------------------------------------------------------------------------------------------------
# Licensed under the MIT LICENSE. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------------------------------
"""Simulation utility functions."""


def trade_tkn_to_ema(
    bnt_trading_liquidity: Decimal,
    tkn_trading_liquidity: Decimal,
    trading_fee: Decimal,
    network_fee: Decimal,
    future_ema: Decimal,
) -> Decimal:
    """
    Outputs the tkn_amt that should be traded to force the ema and the spot price together on a given pool.
    """
    a = bnt_trading_liquidity
    b = tkn_trading_liquidity
    d = trading_fee
    e = network_fee
    f = future_ema
    tkn_amt = (
        (a * d * (Decimal("1") - e) - Decimal("2") * f * b)
        + (
            a
            * (
                Decimal("4") * f * b * (Decimal("1") - d * (Decimal("1") - e))
                + a * d ** Decimal("2") * (Decimal("1") - e) ** Decimal("2")
            )
        )
        ** (Decimal("1") / Decimal("2"))
    ) / (Decimal("2") * f)
    return tkn_amt


def trade_bnt_to_ema(
    bnt_trading_liquidity,
    tkn_trading_liquidity,
    trading_fee,
    network_fee,
    future_ema,
):
    """
    Analyze the state of any pool and create a swap that drives the ema and the spot price together.
    """

    a = bnt_trading_liquidity
    b = tkn_trading_liquidity
    d = trading_fee
    e = network_fee
    f = future_ema
    x = (
        -Decimal("2") * a
        + b * d * f
        + (
            (Decimal("2") * a - b * d * f) ** Decimal("2")
            - Decimal("4") * a * (a - b * f)
        )
        ** (Decimal("1") / Decimal("2"))
    ) / Decimal("2")
    a_recursion = (
        a * (a + x) + d * (Decimal("1") - e) * (a * x + x ** Decimal("2"))
    ) / (a + d * x)
    b_recursion = b * (a + d * x) / (a + x)
    n = 0
    p = Decimal("0.001")
    while a_recursion / b_recursion < f:
        n += 1
        p += Decimal("0.0001")
        x += x * (f**p - (a_recursion / b_recursion) ** p) / f
        a_recursion = (
            a * (a + x) + d * (Decimal("1") - e) * (a * x + x ** Decimal("2"))
        ) / (a + d * x)
        b_recursion = b * (a + d * x) / (a + x)
        if n > 20000:
            break
    bnt_amt = x
    return bnt_amt


def process_arbitrage_trade(
    tkn_name: str,
    tkn_token_virtual_balance: Decimal,
    bnt_virtual_balance: Decimal,
    bnt_trading_liquidity: Decimal,
    tkn_trading_liquidity: Decimal,
    trading_fee: Decimal,
    user_tkn: Decimal,
    user_bnt: Decimal,
) -> Tuple[Decimal, str, str, bool]:
    """
    Computes the appropriate arbitrage trade on the token_name pool.
    """
    a = bnt_trading_liquidity
    b = tkn_trading_liquidity
    m = trading_fee
    p = bnt_virtual_balance
    q = tkn_token_virtual_balance

    bnt_trade_amt = (
        -Decimal("2") * a * q
        + b * m * p
        + (
            (Decimal("2") * a * q - b * m * p) ** Decimal("2")
            - Decimal("4") * a * q * (a * q - b * p)
        )
        ** (Decimal("1") / Decimal("2"))
    ) / (Decimal("2") * q)

    tkn_trade_amt = (
        -Decimal("2") * b * p
        + a * m * q
        + (
            (Decimal("2") * b * p - a * m * q) ** Decimal("2")
            - Decimal("4") * b * p * (b * p - a * q)
        )
        ** (Decimal("1") / Decimal("2"))
    ) / (Decimal("2") * p)

    if bnt_trade_amt > 0:
        source_token = "bnt"
        target_token = tkn_name
        trade_amt = bnt_trade_amt
        user_capability = user_bnt > bnt_trade_amt
        return trade_amt, source_token, target_token, user_capability

    elif tkn_trade_amt > 0:
        source_token = tkn_name
        target_token = "bnt"
        trade_amt = tkn_trade_amt
        user_capability = user_tkn > tkn_trade_amt
        return trade_amt, source_token, target_token, user_capability



from bancor_research.bancor_emulator.solidity import address, revert
from bancor_research.bancor_emulator.utils import contract

'''
 * @dev this contract provides support and utilities for contract ownership
'''
class Owned(contract):
    '''
     * @dev initializes the contract
    '''
    def __init__(self) -> None:
        contract.__init__(self)

        self._owner = address(0);
        self._newOwner = address(0);

        self._setOwnership(self.msg_sender);

    '''
     * @inheritdoc IOwned
    '''
    def owner(self) -> (any):
        return self._owner;

    '''
     * @inheritdoc IOwned
    '''
    def transferOwnership(self, ownerCandidate) -> None:
        if (ownerCandidate == self._owner):
            revert("SameOwner");

        self._newOwner = ownerCandidate;

    '''
     * @inheritdoc IOwned
    '''
    def acceptOwnership(self) -> None:
        if (self.msg_sender != self._newOwner):
            revert("AccessDenied");

        self._setOwnership(self._newOwner);

    '''
     * @dev returns the address of the new owner candidate
    '''
    def newOwner(self) -> (any):
        return self._newOwner;

    '''
     * @dev sets the new owner internally
    '''
    def _setOwnership(self, ownerCandidate) -> None:
        prevOwner = self._owner;

        self._owner = ownerCandidate;
        self._newOwner = address(0);



from bancor_research.bancor_emulator.solidity import uint, uint32, uint256, block

from bancor_research.bancor_emulator.AutoCompoundingRewards import AutoCompoundingRewards
from bancor_research.bancor_emulator.BancorNetwork          import BancorNetwork         
from bancor_research.bancor_emulator.BancorNetworkInfo      import BancorNetworkInfo     
from bancor_research.bancor_emulator.BNTPool                import BNTPool               
from bancor_research.bancor_emulator.Constants              import PPM_RESOLUTION        
from bancor_research.bancor_emulator.ERC20                  import ERC20                 
from bancor_research.bancor_emulator.NetworkSettings        import NetworkSettings       
from bancor_research.bancor_emulator.PendingWithdrawals     import PendingWithdrawals    
from bancor_research.bancor_emulator.PoolCollection         import PoolCollection        
from bancor_research.bancor_emulator.PoolToken              import PoolToken             
from bancor_research.bancor_emulator.PoolTokenFactory       import PoolTokenFactory      
from bancor_research.bancor_emulator.ReserveToken           import ReserveToken          
from bancor_research.bancor_emulator.StandardRewards        import StandardRewards       
from bancor_research.bancor_emulator.TokenGovernance        import TokenGovernance       
from bancor_research.bancor_emulator.Vault                  import Vault                 

from bancor_research import DEFAULT, Decimal, DataFrame, PandasDataFrame, read_price_feeds

def toPPM(percent: str):
    return uint32(Decimal(str(PPM_RESOLUTION)) * Decimal(percent[:-1]) / 100)

def toWei(amount: str, decimals: int):
    return uint256(Decimal(amount) * 10 ** decimals)

def fromWei(amount: uint, decimals: int):
    return Decimal(str(amount)) / 10 ** decimals

def fromFraction(n: uint, d: uint):
    return Decimal('nan') if n == d == 0 else Decimal(str(n)) / Decimal(str(d))

def userAmount(token: ERC20, userId: str, amount: str):
    if (amount.endswith('%')):
        n, d = Decimal(amount[:-1]).as_integer_ratio()
        return token.balanceOf(userId) * n / (d * 100)
    return toWei(amount, token.decimals())

def updateBlock(timestamp):
    if block.timestamp < timestamp:
        block.timestamp = timestamp
        block.number += 1
    else:
        assert block.timestamp == timestamp

class BancorDapp:
    def __init__(
        self,
        timestamp: int = DEFAULT.TIMESTAMP,
        bnt_min_liquidity: str = DEFAULT.BNT_MIN_LIQUIDITY,
        withdrawal_fee: str = DEFAULT.WITHDRAWAL_FEE,
        cooldown_time: int = DEFAULT.COOLDOWN_TIME,
        network_fee: str = DEFAULT.NETWORK_FEE,
        whitelisted_tokens = DEFAULT.WHITELIST,
        price_feeds_path: str = DEFAULT.PRICE_FEEDS_PATH,
        price_feeds: PandasDataFrame = DEFAULT.PRICE_FEEDS,
        log_state: bool = True,
    ):
        block.timestamp = 0
        block.number = 0
        updateBlock(timestamp)

        self.bnt   = ReserveToken('bnt'  , 'bnt'  , DEFAULT.DECIMALS)
        self.vbnt  = ReserveToken('vbnt' , 'vbnt' , DEFAULT.DECIMALS)
        self.bnbnt = PoolToken   ('bnbnt', 'bnbnt', DEFAULT.DECIMALS, self.bnt)

        self.bntGovernance      = TokenGovernance(self.bnt)
        self.vbntGovernance     = TokenGovernance(self.vbnt)
        self.networkSettings    = NetworkSettings(self.bnt)
        self.masterVault        = Vault(self.bntGovernance, self.vbntGovernance)
        self.epVault            = Vault(self.bntGovernance, self.vbntGovernance)
        self.erVault            = Vault(self.bntGovernance, self.vbntGovernance)
        self.network            = BancorNetwork(self.bntGovernance, self.vbntGovernance, self.networkSettings, self.masterVault, self.epVault, self.bnbnt)
        self.bntPool            = BNTPool(self.network, self.bntGovernance, self.vbntGovernance, self.networkSettings, self.masterVault, self.bnbnt)
        self.pendingWithdrawals = PendingWithdrawals(self.network, self.bnt, self.bntPool)
        self.poolTokenFactory   = PoolTokenFactory()
        self.poolMigrator       = None
        self.poolCollection     = PoolCollection(self.network, self.bnt, self.networkSettings, self.masterVault, self.bntPool, self.epVault, self.poolTokenFactory, self.poolMigrator)
        self.standardRewards    = StandardRewards(self.network, self.networkSettings, self.bntGovernance, self.vbnt, self.bntPool)
        self.compoundRewards    = AutoCompoundingRewards(self.network, self.networkSettings, self.bnt, self.bntPool, self.erVault)
        self.networkInfo        = BancorNetworkInfo(self.network, self.bntGovernance, self.vbntGovernance, self.networkSettings, self.masterVault, self.epVault, self.erVault, self.bntPool, self.pendingWithdrawals, self.poolMigrator)

        self.networkSettings.initialize()
        self.network.initialize(self.bntPool, self.pendingWithdrawals, self.poolMigrator)
        self.bntPool.initialize()
        self.pendingWithdrawals.initialize()
        self.poolTokenFactory.initialize()
        self.standardRewards.initialize()
        self.compoundRewards.initialize()
        self.networkInfo.initialize()

        self.networkSettings.setWithdrawalFeePPM(toPPM(withdrawal_fee))
        self.networkSettings.setMinLiquidityForTrading(toWei(bnt_min_liquidity, self.bnt.decimals()))

        self.pendingWithdrawals.setLockDuration(cooldown_time)

        self.poolCollection.setNetworkFeePPM(toPPM(network_fee))

        self.network.registerPoolCollection(self.poolCollection)

        self.reserveTokens = {self.bnt.symbol(): self.bnt}
        self.poolTokens = {self.bnt.symbol(): self.bnbnt}
        for tkn_name, pool_params in whitelisted_tokens.items():
            tkn = ReserveToken(tkn_name, tkn_name, pool_params['decimals'])
            tkn.issue(self.epVault, toWei(pool_params['ep_vault_balance'], tkn.decimals()))
            self.networkSettings.addTokenToWhitelist(tkn)
            self.networkSettings.setFundingLimit(tkn, toWei(pool_params['bnt_funding_limit'], self.bnt.decimals()))
            self.network.createPools([tkn], self.poolCollection)
            self.poolCollection.setTradingFeePPM(tkn, toPPM(pool_params['trading_fee']))
            self.reserveTokens[tkn_name] = tkn
            self.poolTokens[tkn_name] = self.network.collectionByPool(tkn).poolToken(tkn)

        self.price_feeds = price_feeds if price_feeds is not None else read_price_feeds(price_feeds_path)

    def deposit(
        self,
        tkn_name: str,
        tkn_amt: str,
        user_name: str,
        timestamp: int = 0,
        action_name="deposit",
    ):
        updateBlock(timestamp)
        tkn = self.reserveTokens[tkn_name]
        amt = userAmount(tkn, user_name, tkn_amt)
        tkn.connect(user_name).approve(self.network, amt)
        return self.network.connect(user_name).deposit(tkn, amt)

    def trade(
        self,
        tkn_amt: str,
        source_token: str,
        target_token: str,
        user_name: str,
        timestamp: int,
        transaction_type: str = "trade",
    ):
        updateBlock(timestamp)
        src_tkn = self.reserveTokens[source_token]
        trg_tkn = self.reserveTokens[target_token]
        src_amt = userAmount(src_tkn, user_name, tkn_amt)
        src_tkn.connect(user_name).approve(self.network, src_amt)
        return self.network.connect(user_name).tradeBySourceAmount(src_tkn, trg_tkn, src_amt, 1, uint256.max, user_name)

    def begin_cooldown_by_rtkn(
        self,
        tkn_amt: str,
        tkn_name: str,
        user_name: str,
        timestamp: int = 0,
        action_name: str = "begin cooldown by reserve token amount",
    ):
        updateBlock(timestamp)
        tkn = self.poolTokens[tkn_name]
        amt = self.networkInfo.underlyingToPoolToken(self.reserveTokens[tkn_name], userAmount(self.reserveTokens[tkn_name], user_name, tkn_amt))
        tkn.connect(user_name).approve(self.network, amt)
        if tkn is self.bnbnt: self.vbnt.connect(user_name).approve(self.network, amt)
        return self.network.connect(user_name).initWithdrawal(tkn, amt)

    def begin_cooldown_by_ptkn(
        self,
        tkn_amt: str,
        tkn_name: str,
        user_name: str,
        timestamp: int = 0,
        action_name: str = "begin cooldown by pool token amount",
    ):
        updateBlock(timestamp)
        tkn = self.poolTokens[tkn_name]
        amt = userAmount(tkn, user_name, tkn_amt)
        tkn.connect(user_name).approve(self.network, amt)
        if tkn is self.bnbnt: self.vbnt.connect(user_name).approve(self.network, amt)
        return self.network.connect(user_name).initWithdrawal(tkn, amt)

    def withdraw(
        self,
        user_name: str,
        id_number: int,
        timestamp: int = 0,
        transaction_type: str = "withdraw",
    ):
        updateBlock(timestamp)
        return self.network.connect(user_name).withdraw(id_number)

    def burn_pool_tokens(
        self,
        tkn_name: str,
        tkn_amt: str,
        user_name: str,
        timestamp: int = 0,
        transaction_type: str = "burnPoolTokenTKN",
    ):
        updateBlock(timestamp)
        tkn = self.poolTokens[tkn_name]
        amt = userAmount(tkn, user_name, tkn_amt)
        tkn.connect(user_name).burn(amt)

    def create_standard_rewards_program(
        self,
        tkn_name: str,
        rewards_amt: str,
        start_time: int,
        end_time: int,
        timestamp: int = 0,
        transaction_type="create_standard_rewards_program",
    ):
        updateBlock(timestamp)
        tkn = self.reserveTokens[tkn_name]
        amt = toWei(rewards_amt, self.bnt.decimals())
        return self.standardRewards.createProgram(tkn, amt, start_time, end_time)

    def join_standard_rewards_program(
        self,
        tkn_name: str,
        tkn_amt: str,
        user_name: str,
        program_id: int,
        timestamp: int = 0,
        transaction_type="join_standard_rewards_program",
    ):
        updateBlock(timestamp)
        tkn = self.poolTokens[tkn_name]
        amt = userAmount(tkn, user_name, tkn_amt)
        tkn.connect(user_name).approve(self.standardRewards, amt)
        return self.standardRewards.connect(user_name).join(program_id, amt)

    def leave_standard_rewards_program(
        self,
        tkn_name: str,
        tkn_amt: str,
        user_name: str,
        program_id: int,
        timestamp: int = 0,
        transaction_type="leave_standard_rewards_program",
    ):
        updateBlock(timestamp)
        tkn = self.poolTokens[tkn_name]
        amt = userAmount(tkn, user_name, tkn_amt)
        return self.standardRewards.connect(user_name).leave(program_id, amt)

    def claim_standard_rewards(
        self,
        user_name: str,
        program_ids: list,
        timestamp: int = 0,
        transaction_type: str = "claim_standard_rewards",
    ):
        updateBlock(timestamp)
        return self.standardRewards.connect(user_name).claimRewards(program_ids)

    def create_flat_ac_rewards_program(
        self,
        tkn_name: str,
        user_name: str,
        total_rewards: str,
        start_time: int,
        total_duration: int,
        timestamp: int = 0,
        transaction_type: str = "create flat autocompounding rewards program",
    ):
        updateBlock(timestamp)
        tkn = self.reserveTokens[tkn_name]
        amt = userAmount(tkn, user_name, total_rewards)
        self.poolTokens[tkn_name].connect(user_name).transfer(self.erVault, self.networkInfo.underlyingToPoolToken(tkn, amt))
        return self.compoundRewards.createFlatProgram(tkn, amt, start_time, start_time + total_duration)

    def create_exp_ac_rewards_program(
        self,
        tkn_name: str,
        user_name: str,
        total_rewards: str,
        start_time: int,
        half_life: int,
        timestamp: int = 0,
        transaction_type: str = "create exp autocompounding rewards program",
    ):
        updateBlock(timestamp)
        tkn = self.reserveTokens[tkn_name]
        amt = userAmount(tkn, user_name, total_rewards)
        self.poolTokens[tkn_name].connect(user_name).transfer(self.erVault, self.networkInfo.underlyingToPoolToken(tkn, amt))
        return self.compoundRewards.createExpDecayProgram(tkn, amt, start_time, half_life)

    def process_ac_rewards_program(
        self,
        tkn_name: str,
        timestamp: int = 0,
        transaction_type: str = "process autocompounding rewards program",
    ):
        updateBlock(timestamp)
        tkn = self.reserveTokens[tkn_name]
        return self.compoundRewards.processRewards(tkn)

    def terminate_ac_rewards_program(
        self,
        tkn_name: str,
        timestamp: int = 0,
        transaction_type: str = "terminate autocompounding rewards program",
    ):
        updateBlock(timestamp)
        tkn = self.reserveTokens[tkn_name]
        return self.compoundRewards.terminateProgram(tkn)

    def set_user_balance(
        self,
        user_name: str,
        tkn_name: str,
        tkn_amt: str,
        timestamp: int = 0,
        transaction_type: str = "set user balance",
    ):
        updateBlock(timestamp)
        tkn = self.reserveTokens[tkn_name]
        balance = tkn.balanceOf(user_name)
        amount = toWei(tkn_amt, tkn.decimals())
        if tkn is self.bnt:
            if amount > balance:
                self.bntGovernance.mint(user_name, amount - balance)
            elif balance > amount:
                self.bntGovernance.burn(user_name, balance - amount)
        else:
            if amount > balance:
                tkn.issue(user_name, amount - balance)
            elif balance > amount:
                tkn.destroy(user_name, balance - amount)

    def set_trading_fee(
        self,
        tkn_name: str,
        percent: str,
        timestamp: int = 0,
        transaction_type: str = "set trading fee",
    ):
        updateBlock(timestamp)
        self.poolCollection.setTradingFeePPM(self.reserveTokens[tkn_name], toPPM(percent))

    def set_network_fee(
        self,
        tkn_name: str,
        percent: str,
        timestamp: int = 0,
        transaction_type: str = "set network fee",
    ):
        updateBlock(timestamp)
        self.poolCollection.setNetworkFeePPM(toPPM(percent))

    def set_withdrawal_fee(
        self,
        tkn_name: str,
        percent: str,
        timestamp: int = 0,
        transaction_type: str = "set withdrawal fee",
    ):
        updateBlock(timestamp)
        self.networkSettings.setWithdrawalFeePPM(toPPM(percent))

    def set_bnt_funding_limit(
        self,
        tkn_name: str,
        amount: str,
        timestamp: int = 0,
        transaction_type: str = "set bnt funding limit",
    ):
        updateBlock(timestamp)
        tkn = self.reserveTokens[tkn_name]
        amt = toWei(amount, self.bnt.decimals())
        self.networkSettings.setFundingLimit(tkn, amt)

    def enable_trading(
        self,
        tkn_name: str,
        tkn_price: str = "",
        bnt_price: str = "",
        timestamp: int = 0,
        transaction_type: str = "enableTrading",
    ) -> None:
        updateBlock(timestamp)
        tknPrice = Decimal(tkn_price) if tkn_price else self.price_feeds.at[timestamp, tkn_name]
        bntPrice = Decimal(bnt_price) if bnt_price else self.price_feeds.at[timestamp, self.bnt.symbol()]
        while tknPrice != int(tknPrice) or bntPrice != int(bntPrice):
            tknPrice *= 10
            bntPrice *= 10
        self.poolCollection.enableTrading(self.reserveTokens[tkn_name], tknPrice, bntPrice)

    def create_user(self, user_name: str, timestamp: int = 0):
        pass

    def describe(self, decimals: int = -1):
        table = {}

        reserveTokens = list(self.reserveTokens.values())
        poolTokens = list(self.poolTokens.values())

        # Iterate all tokens
        for token in reserveTokens + poolTokens + [self.vbnt]:
            table[token.symbol()] = {}
            for account in [user for user in token._balances.keys() if type(user) is str]:
                table[token.symbol()][tuple([1, 'Account', account])] = fromWei(token.balanceOf(account), token.decimals())

        # Iterate all reserve tokens except bnt
        for token in [reserveToken for reserveToken in reserveTokens if reserveToken is not self.bnt]:
            stakedBalance = self.networkInfo.stakedBalance(token)
            tradingLiquidity = self.networkInfo.tradingLiquidity(token)
            currentPoolFunding = self.bntPool.currentPoolFunding(token)

            spotRateN       = self.poolCollection._poolData[token].liquidity.bntTradingLiquidity
            spotRateD       = self.poolCollection._poolData[token].liquidity.baseTokenTradingLiquidity
            averageRateN    = self.poolCollection._poolData[token].averageRates.rate.n
            averageRateD    = self.poolCollection._poolData[token].averageRates.rate.d
            averageInvRateN = self.poolCollection._poolData[token].averageRates.invRate.n
            averageInvRateD = self.poolCollection._poolData[token].averageRates.invRate.d

            table[token.symbol()][tuple([2, 'Pool', 'a: TKN Staked Balance'   ])] = fromWei(stakedBalance, token.decimals())
            table[token.symbol()][tuple([2, 'Pool', 'b: TKN Trading Liquidity'])] = fromWei(tradingLiquidity.baseTokenTradingLiquidity, token.decimals())
            table[token.symbol()][tuple([2, 'Pool', 'c: BNT Trading Liquidity'])] = fromWei(tradingLiquidity.bntTradingLiquidity, self.bnt.decimals())
            table[token.symbol()][tuple([2, 'Pool', 'd: BNT Current Funding'  ])] = fromWei(currentPoolFunding, self.bnt.decimals())
            table[token.symbol()][tuple([2, 'Pool', 'e: Spot Rate'            ])] = fromFraction(spotRateN      , spotRateD      )
            table[token.symbol()][tuple([2, 'Pool', 'f: Average Rate'         ])] = fromFraction(averageRateN   , averageRateD   )
            table[token.symbol()][tuple([2, 'Pool', 'g: Average Inverse Rate' ])] = fromFraction(averageInvRateN, averageInvRateD)

        # Iterate all reserve tokens
        for token in reserveTokens:
            table[token.symbol()][tuple([3, 'Network', 'Master Vault'    ])] = fromWei(token.balanceOf(self.masterVault), token.decimals())
            table[token.symbol()][tuple([3, 'Network', 'Protection Vault'])] = fromWei(token.balanceOf(self.epVault    ), token.decimals())

        # Iterate all pool tokens
        for token in poolTokens:
            table[token.symbol()][tuple([3, 'Network', 'Rewards Vault'  ])] = fromWei(token.balanceOf(self.erVault), token.decimals())
            table[token.symbol()][tuple([3, 'Network', 'Protocol Equity'])] = fromWei(token.balanceOf(self.bntPool), token.decimals())

        # Iterate only bnt
        table[self.bnt.symbol()][tuple([3, 'Network', 'BNT Staked Balance'])] = fromWei(self.bntPool.stakedBalance(), self.bnt.decimals())

        df = DataFrame(table).fillna(0).sort_index()
        return df.applymap(lambda x: round(x, decimals)) if decimals >= 0 else df



# coding=utf-8
# --------------------------------------------------------------------------------------------------------------------
# Licensed under the MIT LICENSE. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------------------------------
"""
Script used to perform a parameter sweep.

Note:
    This version of the model has a BatchRunner at the bottom. This
    is for collecting data on parameter sweeps. It is not meant to
    be run with run.py, since run.py starts up a server for visualization, which
    isn't necessary for the BatchRunner. To run a parameter sweep, call
    batch_run.py in the command line.

    The BatchRunner is set up to collect step by step data of the model. It does
    this by collecting the DataCollector object in a model_reporter (i.e. the
    DataCollector is collecting itself every step).

    The end result of the batch run will be a CSV file created in the same
    directory from which Python was run. The CSV file will contain the data from
    every step of every run.
"""
import os

from bancor_research import DEFAULT, Decimal, DataFrame

trading_fee = Decimal(DEFAULT.TRADING_FEE[:-1]) / 100

# Add any desired parameters to sweep below.
br_params = {
    "trading_fee": [
        trading_fee - Decimal("0.001"),
        trading_fee,
        trading_fee + Decimal("0.001"),
    ]
}

if __name__ == "__main__":
    data = mesa.batch_run(
        model_cls=BancorSimulation,
        parameters=br_params,
        data_collection_period=1,
        max_steps=SIMULATION_MAX_STEPS,
    )
    br_df = DataFrame(data)

    overwrite_dir = True
    os.makedirs(SIMULATION_OUTPUT_PATH, exist_ok=overwrite_dir)

    br_df.to_csv(SIMULATION_OUTPUT_PATH)



from bancor_research.bancor_emulator.solidity import uint, uint32, block

'''
 * @dev this contract abstracts the block timestamp in order to allow for more flexible control in tests
'''
class Time:
    '''
     * @dev returns the current time
    '''
    def _time(self) -> (uint):
        return uint32(block.timestamp);



from bancor_research.bancor_emulator.ERC20 import ERC20
from bancor_research.bancor_emulator.ERC20Burnable import ERC20Burnable
from bancor_research.bancor_emulator.Owned import Owned

'''
 * @dev Pool Token contract
'''
class PoolToken(ERC20Burnable, Owned):
    '''
     * @dev initializes a new PoolToken contract
    '''
    def __init__(self,
        name,
        symbol,
        initDecimals,
        initReserveToken
    ) -> None:
        ERC20.__init__(self, name, symbol)
        Owned.__init__(self)

        self._decimals = initDecimals;
        self._reserveToken = initReserveToken;

    '''
     * @inheritdoc IVersioned
    '''
    def version(self) -> (int):
        return 1;

    '''
     * @dev returns the number of decimals used to get its user representation
    '''
    def decimals(self):
        return self._decimals;

    '''
     * @inheritdoc IPoolToken
    '''
    def reserveToken(self):
        return self._reserveToken;

    '''
     * @inheritdoc IPoolToken
    '''
    def mint(self, recipient, amount):
        self._mint(recipient, amount);



from bancor_research.bancor_emulator.ERC20 import ERC20

'''
 * @dev this is an adapted clone of the OZ's ERC20Burnable extension which is unfortunately required so that it can be
 * explicitly specified in interfaces via our new IERC20Burnable interface.
 *
 * We have also removed the explicit use of Context and updated the code to our style.
'''
class ERC20Burnable(ERC20):
    '''
     * @inheritdoc IERC20Burnable
    '''
    def burn(self, amount) -> None:
        self._burn(self.msg_sender, amount);

    '''
     * @inheritdoc IERC20Burnable
    '''
    def burnFrom(self, recipient, amount) -> None:
        self._approve(recipient, self.msg_sender, self.allowance(recipient, self.msg_sender) - amount);
        self._burn(recipient, amount);



from . uint import uint
from collections import defaultdict

def address(obj):
    return obj

def payable(obj):
    return obj

mapping = defaultdict

for size in uint.sizes:
    exec('def uint{}(data = 0): return uint({}, data)'.format(size, size))
    exec('uint{}.max = uint{}(2 ** {} - 1)'.format(size, size, size))



from bancor_research.bancor_emulator.solidity import uint, uint256, unchecked, mapping
from bancor_research.bancor_emulator.utils import contract

'''
 * @dev Implementation of the {IERC20} interface.
 *
 * This implementation is agnostic to the way tokens are created. This means
 * that a supply mechanism has to be added in a derived contract using {_mint}.
 * For a generic mechanism see {ERC20PresetMinterPauser}.
 *
 * TIP: For a detailed writeup see our guide
 * https://forum.zeppelin.solutions/t/how-to-implement-erc20-supply-mechanisms/226[How
 * to implement supply mechanisms].
 *
 * We have followed general OpenZeppelin Contracts guidelines: functions revert
 * instead returning `false` on failure. This behavior is nonetheless
 * conventional and does not conflict with the expectations of ERC20
 * applications.
 *
 * Additionally, an {Approval} event is emitted on calls to {transferFrom}.
 * This allows applications to reconstruct the allowance for all accounts just
 * by listening to said events. Other implementations of the EIP may not emit
 * these events, as it isn't required by the specification.
 *
 * Finally, the non-standard {decreaseAllowance} and {increaseAllowance * functions have been added to mitigate the well-known issues around setting
 * allowances. See {IERC20-approve}.
'''
class ERC20(contract):
    '''
     * @dev Sets the values for {name} and {symbol}.
     *
     * The default value of {decimals} is 18. To select a different value for
     * {decimals} you should overload it.
     *
     * All two of these values are immutable: they can only be set once during
     * construction.
    '''
    def __init__(self, name_: str, symbol_: str) -> None:
        contract.__init__(self)

        self._name = name_;
        self._symbol = symbol_;
        self._balances = mapping(lambda: uint256());
        self._allowances = mapping(lambda: mapping(lambda: uint256()));
        self._totalSupply = uint256();

    '''
     * @dev Returns the name of the token.
    '''
    def name(self) -> (str):
        return self._name;

    '''
     * @dev Returns the symbol of the token, usually a shorter version of the
     * name.
    '''
    def symbol(self) -> (str):
        return self._symbol;

    '''
     * @dev Returns the number of decimals used to get its user representation.
     * For example, if `decimals` equals `2`, a balance of `505` tokens should
     * be displayed to a user as `5.05` (`505 / 10 ** 2`).
     *
     * Tokens usually opt for a value of 18, imitating the relationship between
     * Ether and Wei. This is the value {ERC20} uses, unless this def is
     * overridden;
     *
     * NOTE: This information is only used for _display_ purposes: it in
     * no way affects any of the arithmetic of the contract, including
     * {IERC20-balanceOf} and {IERC20-transfer}.
    '''
    def decimals(self) -> (int):
        return 18;

    '''
     * @dev See {IERC20-totalSupply}.
    '''
    def totalSupply(self) -> (uint):
        return self._totalSupply.clone();

    '''
     * @dev See {IERC20-balanceOf}.
    '''
    def balanceOf(self, account) -> (uint):
        return self._balances[account].clone();

    '''
     * @dev See {IERC20-transfer}.
     *
     * Requirements:
     *
     * - `to` cannot be the zero address.
     * - the caller must have a balance of at least `amount`.
    '''
    def transfer(self, to, amount) -> (bool):
        owner = self.msg_sender;
        self._transfer(owner, to, amount);
        return True;

    '''
     * @dev See {IERC20-allowance}.
    '''
    def allowance(self, owner, spender) -> (uint):
        return self._allowances[owner][spender].clone();

    '''
     * @dev See {IERC20-approve}.
     *
     * NOTE: If `amount` is the maximum `uint256`, the allowance is not updated on
     * `transferFrom`. This is semantically equivalent to an infinite approval.
     *
     * Requirements:
     *
     * - `spender` cannot be the zero address.
    '''
    def approve(self, spender, amount) -> (bool):
        owner = self.msg_sender;
        self._approve(owner, spender, amount);
        return True;

    '''
     * @dev See {IERC20-transferFrom}.
     *
     * Emits an {Approval} event indicating the updated allowance. This is not
     * required by the EIP. See the note at the beginning of {ERC20}.
     *
     * NOTE: Does not update the allowance if the current allowance
     * is the maximum `uint256`.
     *
     * Requirements:
     *
     * - `from` and `to` cannot be the zero address.
     * - `from` must have a balance of at least `amount`.
     * - the caller must have allowance for ``from``'s tokens of at least
     * `amount`.
    '''
    def transferFrom(self,
        from_,
        to,
        amount
    ) -> (bool):
        spender = self.msg_sender;
        self._spendAllowance(from_, spender, amount);
        self._transfer(from_, to, amount);
        return True;

    '''
     * @dev Atomically increases the allowance granted to `spender` by the caller.
     *
     * This is an alternative to {approve} that can be used as a mitigation for
     * problems described in {IERC20-approve}.
     *
     * Emits an {Approval} event indicating the updated allowance.
     *
     * Requirements:
     *
     * - `spender` cannot be the zero address.
    '''
    def increaseAllowance(self, spender, addedValue) -> (bool):
        owner = self.msg_sender;
        self._approve(owner, spender, self._allowances[owner][spender] + addedValue);
        return True;

    '''
     * @dev Atomically decreases the allowance granted to `spender` by the caller.
     *
     * This is an alternative to {approve} that can be used as a mitigation for
     * problems described in {IERC20-approve}.
     *
     * Emits an {Approval} event indicating the updated allowance.
     *
     * Requirements:
     *
     * - `spender` cannot be the zero address.
     * - `spender` must have allowance for the caller of at least
     * `subtractedValue`.
    '''
    def decreaseAllowance(self, spender, subtractedValue) -> (bool):
        owner = self.msg_sender;
        currentAllowance = self._allowances[owner][spender];
        assert currentAllowance >= subtractedValue, "ERC20: decreased allowance below zero";
        unchecked.begin()
        self._approve(owner, spender, currentAllowance - subtractedValue);
        unchecked.end()

        return True;

    '''
     * @dev Moves `amount` of tokens from `sender` to `recipient`.
     *
     * This internal def is equivalent to {transfer}, and can be used to
     * e.g. implement automatic token fees, slashing mechanisms, etc.
     *
     * Emits a {Transfer} event.
     *
     * Requirements:
     *
     * - `from` cannot be the zero address.
     * - `to` cannot be the zero address.
     * - `from` must have a balance of at least `amount`.
    '''
    def _transfer(self,
        from_,
        to,
        amount
    ):
        fromBalance = self._balances[from_];
        assert fromBalance >= amount, "ERC20: transfer amount exceeds balance";
        unchecked.begin()
        self._balances[from_] = fromBalance - amount;
        unchecked.end()
        self._balances[to] += amount;

    ''' @dev Creates `amount` tokens and assigns them to `account`, increasing
     * the total supply.
     *
     * Emits a {Transfer} event with `from` set to the zero address.
     *
     * Requirements:
     *
     * - `account` cannot be the zero address.
    '''
    def _mint(self, account, amount):
        self._totalSupply += amount;
        self._balances[account] += amount;

    '''
     * @dev Destroys `amount` tokens from `account`, reducing the
     * total supply.
     *
     * Emits a {Transfer} event with `to` set to the zero address.
     *
     * Requirements:
     *
     * - `account` cannot be the zero address.
     * - `account` must have at least `amount` tokens.
    '''
    def _burn(self, account, amount):
        accountBalance = self._balances[account];
        assert accountBalance >= amount, "ERC20: burn amount exceeds balance";
        unchecked.begin()
        self._balances[account] = accountBalance - amount;
        unchecked.end()
        self._totalSupply -= amount;

    '''
     * @dev Sets `amount` as the allowance of `spender` over the `owner` s tokens.
     *
     * This internal def is equivalent to `approve`, and can be used to
     * e.g. set automatic allowances for certain subsystems, etc.
     *
     * Emits an {Approval} event.
     *
     * Requirements:
     *
     * - `owner` cannot be the zero address.
     * - `spender` cannot be the zero address.
    '''
    def _approve(self,
        owner,
        spender,
        amount
    ):
        self._allowances[owner][spender] = uint256(amount);

    '''
     * @dev Spend `amount` form the allowance of `owner` toward `spender`.
     *
     * Does not update the allowance amount in case of infinite allowance.
     * Revert if not enough allowance is available.
     *
     * Might emit an {Approval} event.
    '''
    def _spendAllowance(self,
        owner,
        spender,
        amount
    ):
        currentAllowance = self.allowance(owner, spender);
        if (currentAllowance != uint256.max):
            assert currentAllowance >= amount, "ERC20: insufficient allowance";
            unchecked.begin()
            self._approve(owner, spender, currentAllowance - amount);
            unchecked.end()



"""Logging functions."""
import logging
import sys
import time


def get_logger(name):
    """Get the logger with the associated name.

    Args:
        name (str): Name of the logger to get.

    Returns:
        The logger object with the associated name.
    """
    logger = logging.getLogger(name)
    if not len(logger.handlers):
        logger.setLevel(logging.DEBUG)
        stdout_handler = logging.StreamHandler(sys.stdout)
        stdout_handler.setLevel(logging.INFO)
        stdout_handler.setFormatter(logging.Formatter("%(message)s"))
        logger.addHandler(stdout_handler)
    return logger


def log_title(logger, title):
    """Log with a title."""
    logger.info("\n" + "*" * (len(title) + 4))
    logger.info("* %s *" % title)
    logger.info("*" * (len(title) + 4))
    logger.info("")


def log_subtitle(logger, title, underline="="):
    """Log with a subtitle."""
    logger.info("")
    logger.info("%s" % title)
    logger.info(underline * len(title))


def time_elapsed(start_time):
    """How much time has elapsed since the search started.

    Args:
        start_time (int): Time when search started.

    Returns:
        str: elapsed time formatted as a string [H:]MM:SS
    """
    time_diff = time.time() - start_time
    # Source: tqdm.std.tqdm.format_interval
    mins, s = divmod(int(time_diff), 60)
    h, m = divmod(mins, 60)
    if h:
        return "{0:d}:{1:02d}:{2:02d}".format(h, m, s)
    else:
        return "{0:02d}:{1:02d}".format(m, s)



from bancor_research.bancor_emulator.solidity import uint, uint128, address, revert
from bancor_research.bancor_emulator.utils import contract, parse

from bancor_research.bancor_emulator import ERC20 as IERC20
from bancor_research.bancor_emulator import TokenGovernance as ITokenGovernance
from bancor_research.bancor_emulator.Vault import Vault as IVault
from bancor_research.bancor_emulator import PoolToken as IPoolToken
from bancor_research.bancor_emulator import PoolCollection as IPoolCollection
from bancor_research.bancor_emulator.PoolCollection import WithdrawalAmounts
from bancor_research.bancor_emulator import BNTPool as IBNTPool
from bancor_research.bancor_emulator import BancorNetwork as IBancorNetwork
from bancor_research.bancor_emulator import NetworkSettings as INetworkSettings
from bancor_research.bancor_emulator import PendingWithdrawals as IPendingWithdrawals

class TradingLiquidity:
    def __init__(self, x = None) -> None:
        self.bntTradingLiquidity = parse(uint128, x, 'bntTradingLiquidity');
        self.baseTokenTradingLiquidity = parse(uint128, x, 'baseTokenTradingLiquidity');

'''
 * @dev Bancor Network Information contract
'''
class BancorNetworkInfo(contract):
    '''
     * @dev a "virtual" constructor that is only used to set immutable state variables
    '''
    def __init__(self,
        initNetwork,
        initBNTGovernance,
        initVBNTGovernance,
        initNetworkSettings,
        initMasterVault,
        initExternalProtectionVault,
        initExternalRewardsVault,
        initBNTPool,
        initPendingWithdrawals,
        initPoolMigrator
    ) -> None:
        contract.__init__(self)

        self._network = initNetwork;
        self._bntGovernance = initBNTGovernance;
        self._bnt = initBNTGovernance.token();
        self._vbntGovernance = initVBNTGovernance;
        self._vbnt = initVBNTGovernance.token();
        self._networkSettings = initNetworkSettings;
        self._masterVault = initMasterVault;
        self._externalProtectionVault = initExternalProtectionVault;
        self._externalRewardsVault = initExternalRewardsVault;
        self._bntPool = initBNTPool;
        self._bntPoolToken = initBNTPool.poolToken();
        self._pendingWithdrawals = initPendingWithdrawals;
        self._poolMigrator = initPoolMigrator;

    '''
     * @dev fully initializes the contract and its parents
    '''
    def initialize(self) -> None:
        self.__BancorNetworkInfo_init();

    # solhint-disable func-name-mixedcase

    '''
     * @dev initializes the contract and its parents
    '''
    def __BancorNetworkInfo_init(self) -> None:
        self.__BancorNetworkInfo_init_unchained();

    '''
     * @dev performs contract-specific initialization
    '''
    def __BancorNetworkInfo_init_unchained(self) -> None:
        pass

    # solhint-enable func-name-mixedcase

    '''
     * @inheritdoc Upgradeable
    '''
    def version(self) -> (int):
        return 2;

    '''
     * @inheritdoc IBancorNetworkInfo
    '''
    def network(self) -> (IBancorNetwork):
        return self._network;

    '''
     * @inheritdoc IBancorNetworkInfo
    '''
    def bnt(self) -> (IERC20):
        return self._bnt;

    '''
     * @inheritdoc IBancorNetworkInfo
    '''
    def bntGovernance(self) -> (ITokenGovernance):
        return self._bntGovernance;

    '''
     * @inheritdoc IBancorNetworkInfo
    '''
    def vbnt(self) -> (IERC20):
        return self._vbnt;

    '''
     * @inheritdoc IBancorNetworkInfo
    '''
    def vbntGovernance(self) -> (ITokenGovernance):
        return self._vbntGovernance;

    '''
     * @inheritdoc IBancorNetworkInfo
    '''
    def networkSettings(self) -> (INetworkSettings):
        return self._networkSettings;

    '''
     * @inheritdoc IBancorNetworkInfo
    '''
    def masterVault(self) -> (IVault):
        return self._masterVault;

    '''
     * @inheritdoc IBancorNetworkInfo
    '''
    def externalProtectionVault(self) -> (IVault):
        return self._externalProtectionVault;

    '''
     * @inheritdoc IBancorNetworkInfo
    '''
    def externalRewardsVault(self) -> (IVault):
        return self._externalRewardsVault;

    '''
     * @inheritdoc IBancorNetworkInfo
    '''
    def bntPool(self) -> (IBNTPool):
        return self._bntPool;

    '''
     * @inheritdoc IBancorNetworkInfo
    '''
    def poolToken(self, pool) -> (IPoolToken):
        return self._bntPoolToken if pool == self._bnt else self._poolCollection(pool).poolToken(pool);

    '''
     * @inheritdoc IBancorNetworkInfo
    '''
    def stakedBalance(self, pool) -> (uint):
        return self._bntPool.stakedBalance() if pool == self._bnt else self._poolCollection(pool).poolLiquidity(pool).stakedBalance;

    '''
     * @inheritdoc IBancorNetworkInfo
    '''
    def tradingLiquidity(self, pool) -> (TradingLiquidity):
        if (pool == self._bnt):
            revert("InvalidParam");

        liquidity = self._poolCollection(pool).poolLiquidity(pool);

        return \
            TradingLiquidity({
                'bntTradingLiquidity': liquidity.bntTradingLiquidity,
                'baseTokenTradingLiquidity': liquidity.baseTokenTradingLiquidity
            });

    '''
     * @inheritdoc IBancorNetworkInfo
    '''
    def tradingFeePPM(self, pool) -> (uint):
        if (pool == self._bnt):
            revert("InvalidParam");

        return self._poolCollection(pool).tradingFeePPM(pool);

    '''
     * @inheritdoc IBancorNetworkInfo
    '''
    def tradingEnabled(self, pool) -> (bool):
        return True if pool == self._bnt else self._poolCollection(pool).tradingEnabled(pool);

    '''
     * @inheritdoc IBancorNetworkInfo
    '''
    def depositingEnabled(self, pool) -> (bool):
        return True if pool == self._bnt else self._poolCollection(pool).depositingEnabled(pool);

    '''
     * @inheritdoc IBancorNetworkInfo
    '''
    def isPoolStable(self, pool) -> (bool):
        return True if pool == self._bnt else self._poolCollection(pool).isPoolStable(pool);

    '''
     * @inheritdoc IBancorNetworkInfo
    '''
    def pendingWithdrawals(self) -> (IPendingWithdrawals):
        return self._pendingWithdrawals;

    '''
     * @inheritdoc IBancorNetworkInfo
    '''
    def poolMigrator(self) -> (any):
        return self._poolMigrator;

    '''
     * @inheritdoc IBancorNetworkInfo
    '''
    def tradeOutputBySourceAmount(self,
        sourceToken,
        targetToken,
        sourceAmount
    ) -> (uint):
        return self._tradeOutputAmount(sourceToken, targetToken, sourceAmount, True);

    '''
     * @inheritdoc IBancorNetworkInfo
    '''
    def tradeInputByTargetAmount(self,
        sourceToken,
        targetToken,
        targetAmount
    ) -> (uint):
        return self._tradeOutputAmount(sourceToken, targetToken, targetAmount, False);

    '''
     * @inheritdoc IBancorNetworkInfo
    '''
    def isReadyForWithdrawal(self, id) -> (bool):
        return self._pendingWithdrawals.isReadyForWithdrawal(id);

    '''
     * @inheritdoc IBancorNetworkInfo
    '''
    def poolTokenToUnderlying(self, pool, poolTokenAmount) -> (uint):
        return \
            self._bntPool.poolTokenToUnderlying(poolTokenAmount) if pool == self._bnt else \
            self._poolCollection(pool).poolTokenToUnderlying(pool, poolTokenAmount);

    '''
     * @inheritdoc IBancorNetworkInfo
    '''
    def underlyingToPoolToken(self, pool, tokenAmount) -> (uint):
        return \
            self._bntPool.underlyingToPoolToken(tokenAmount) if pool == self._bnt else \
            self._poolCollection(pool).underlyingToPoolToken(pool, tokenAmount);

    '''
     * @inheritdoc IBancorNetworkInfo
    '''
    def withdrawalAmounts(self, pool, poolTokenAmount) -> (WithdrawalAmounts):
        if (pool == self._bnt):
            amount = self._bntPool.withdrawalAmount(poolTokenAmount);
            return WithdrawalAmounts({ 'totalAmount': amount, 'baseTokenAmount': 0, 'bntAmount': amount });

        poolCollection = self._poolCollection(pool);
        return poolCollection.withdrawalAmounts(pool, poolTokenAmount);

    '''
     * @dev returns either the source amount or the target amount by providing the source and the target tokens
     * and whether we're interested in the target or the source amount
    '''
    def _tradeOutputAmount(self,
        sourceToken,
        targetToken,
        amount,
        bySourceAmount
    ) -> (uint):
        isSourceBNT = sourceToken == self._bnt;
        isTargetBNT = targetToken == self._bnt;

        # return the trade amount when trading BNT
        if (isSourceBNT or isTargetBNT):
            token = targetToken if isSourceBNT else sourceToken;
            poolCollection = self._poolCollection(token);

            return \
                (
                    poolCollection.tradeOutputAndFeeBySourceAmount(sourceToken, targetToken, amount) if bySourceAmount else
                    poolCollection.tradeInputAndFeeByTargetAmount(sourceToken, targetToken, amount)
                ).amount;

        # return the target amount by simulating double-hop trade from the source token to the target token via BNT
        if (bySourceAmount):
            targetAmount = self._poolCollection(sourceToken) \
                .tradeOutputAndFeeBySourceAmount(sourceToken, address(self._bnt), amount) \
                .amount;

            return \
                self._poolCollection(targetToken) \
                    .tradeOutputAndFeeBySourceAmount(address(self._bnt), targetToken, targetAmount) \
                    .amount;

        # return the source amount by simulating a "reverse" double-hop trade from the source token to the target token
        # via BNT
        requireNetworkAmount = self._poolCollection(targetToken) \
            .tradeInputAndFeeByTargetAmount(address(self._bnt), targetToken, amount) \
            .amount;

        return \
            self._poolCollection(sourceToken) \
                .tradeInputAndFeeByTargetAmount(sourceToken, address(self._bnt), requireNetworkAmount) \
                .amount;

    '''
     * @dev verifies that the specified pool is managed by a valid pool collection and returns it
    '''
    def _poolCollection(self, token) -> (IPoolCollection):
        # verify that the pool is managed by a valid pool collection
        poolCollection = self._network.collectionByPool(token);
        if (address(poolCollection) == address(0)):
            revert("InvalidToken");

        return poolCollection;



from bancor_research.bancor_emulator.solidity import uint

from bancor_research.bancor_emulator.FractionLibrary import Fraction256
from bancor_research.bancor_emulator.MathEx import MathEx

'''
 * @dev This library supports the calculation of staking rewards
'''
class RewardsMath:
    '''
        * @dev returns the amount of rewards distributed on a flat amount ratio
    '''
    def calcFlatRewards(
        totalRewards,
        timeElapsed,
        programDuration
    ) -> (uint):
        # ensures that the function never returns more than the total rewards
        assert(timeElapsed <= programDuration);
        return MathEx.mulDivF(totalRewards, timeElapsed, programDuration);

    '''
        * @dev returns the amount of rewards distributed after a given time period since deployment has elapsed
        *
        * the returned value is calculated as `totalRewards * (1 - 1 / 2 ^ (timeElapsed / halfLife))`
        * note that because the exponentiation function is limited to an input of up to (and excluding)
        * 16 / ln 2, the input value to this function is limited by `timeElapsed / halfLife < 16 / ln 2`
    '''
    def calcExpDecayRewards(
        totalRewards,
        timeElapsed,
        halfLife
    ) -> (uint):
        input = Fraction256({ 'n': timeElapsed, 'd': halfLife });
        output = MathEx.exp2(input);
        return MathEx.mulDivF(totalRewards, output.n - output.d, output.n);



# coding=utf-8
"""CLI commands."""
import click

from bancor_research.bancor_simulator.v3.cli_utils import print_info


@click.group()
def cli():
    """CLI command with no arguments. Does nothing."""
    pass


@click.command()
def info():
    """CLI command with `info` argument. Prints info about the system, bancor3, and dependencies of bancor3."""
    print_info()


cli.add_command(info)



# coding=utf-8
# --------------------------------------------------------------------------
# Licensed under the MIT LICENSE. See License.txt in the project root for license information.
# --------------------------------------------------------------------------

"""
Generalized behavior for random walking, one grid cell at a time.
"""

import mesa


class RandomWalker(mesa.Agent):
    """
    Class implementing random walker methods in a generalized manner.
    """

    grid = None
    x = None
    y = None

    # use a Moore neighborhood
    moore = True

    def __init__(self, unique_id, pos, model, moore=True):
        """
        grid: The MultiGrid object in which the agent lives.
        x: The agent's current x coordinate
        y: The agent's current y coordinate
        moore: If True, may move in all 8 directions.
                Otherwise, only up, down, left, right.
        """
        super().__init__(unique_id, model)
        self.pos = pos
        self.moore = moore

    def random_move(self):
        """
        Step one cell in any allowable direction.
        """
        # Pick the next cell from the adjacent cells.
        next_moves = self.model.grid.get_neighborhood(self.pos, self.moore, True)
        next_move = self.random.choice(next_moves)
        # Now move:
        self.model.grid.move_agent(self, next_move)
